from dataclasses import asdict, make_dataclass
from datetime import timedelta
from pprint import pformat
from typing import Any, Dict, Optional, Tuple, Type

import joblib
import pandas as pd

# from dataclasses_json import DataClassJsonMixin as DataClassJSONMixin
from flytekit import Resources, task, workflow
from flytekit.extras.accelerators import T4
from flytekit.types.file import JoblibSerializedFile
from mashumaro.mixins.json import DataClassJSONMixin
from sklearn.datasets import load_wine
from sklearn.linear_model import LogisticRegression

from pyrovelocity.logging import configure_logging
from pyrovelocity.workflows.configuration import create_dataclass_from_callable

logger = configure_logging("pyrovelocity.workflows.lrwine")

# This is an optional dictionary that can be used to override the
# default types and values inferred from the callable if necessary
# or required. For this example, we provide commented out defaults
# to illustrate the types that are inferred from the callable and
# the ability to override them.
custom_types_defaults: Dict[str, Tuple[Type, Any]] = {
    # "penalty": (str, "l2"),
    # "dual": (bool, False),
    # "tol": (float, 1e-4),
    # "C": (float, 1.0),
    # "fit_intercept": (bool, True),
    # "intercept_scaling": (int, 1),
    "class_weight": (Optional[dict], None),
    "random_state": (Optional[int], None),
    # "solver": (str, "lbfgs"),
    "max_iter": (int, 2000),
    # "multi_class": (str, "auto"),
    # "verbose": (int, 0),
    # "warm_start": (bool, False),
    "n_jobs": (Optional[int], None),
    "l1_ratio": (Optional[float], None),
}

logistic_regression_fields = create_dataclass_from_callable(
    LogisticRegression, custom_types_defaults
)

LogisticRegressionInterface = make_dataclass(
    "LogisticRegressionInterface",
    logistic_regression_fields,
    bases=(DataClassJSONMixin,),
    # TODO: Python 3.12, https://github.com/python/cpython/pull/102104
    # module=__name__,
)
LogisticRegressionInterface.__module__ = __name__


# The following can be used to test dynamic dataclass construction
# with multiple dataclasses of distinct types.
# from sklearn.linear_model import LinearRegression

# linear_regression_custom_types: Dict[str, Tuple[Type,Any]] = {
#     "n_jobs": (Optional[int], None),
# }
# linear_regression_fields = create_dataclass_from_callable(
#     LinearRegression, linear_regression_custom_types
# )

# LinearRegressionInterface = make_dataclass(
#     "LinearRegressionInterface",
#     linear_regression_fields,
#     bases=(DataClassJSONMixin,),
# )
# LinearRegressionInterface.__module__ = __name__


sample_columns = [
    "alcohol",
    "target",
]

sample_data = [
    [13.0, 0],
    [14.0, 1],
    [12.5, 2],
]


@task(
    cache=True,
    cache_version="0.1.0",
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=20),
    container_image="{{.image.gpu.fqn}}:{{.image.gpu.version}}",
    requests=Resources(
        cpu="200m", mem="400Mi", ephemeral_storage="1Gi", gpu="1"
    ),
    accelerator=T4,
)
def get_data() -> pd.DataFrame:
    """
    Get the wine dataset.
    """
    # import time

    # time.sleep(7200)
    return load_wine(as_frame=True).frame


@task(
    cache=False,
    cache_version="0.1.0",
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=10),
    requests=Resources(cpu="200m", mem="400Mi", ephemeral_storage="1Gi"),
)
def process_data(
    data: pd.DataFrame = pd.DataFrame(data=sample_data, columns=sample_columns),
) -> pd.DataFrame:
    """
    Simplify the task from a 3-class to a binary classification problem.
    """
    return data.assign(target=lambda x: x["target"].where(x["target"] == 0, 1))


@task(
    cache=True,
    cache_version="0.1.0",
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=10),
    requests=Resources(cpu="200m", mem="400Mi", ephemeral_storage="1Gi"),
)
def train_model(
    data: pd.DataFrame = pd.DataFrame(data=sample_data, columns=sample_columns),
    logistic_regression: LogisticRegressionInterface = LogisticRegressionInterface(
        max_iter=1200
    ),
) -> JoblibSerializedFile:
    """
    Train a model on the wine dataset.
    """
    features = data.drop("target", axis="columns")
    target = data["target"]
    logger.info(f"{pformat(logistic_regression)}\n\n")
    model = LogisticRegression(**asdict(logistic_regression))
    model_path = "logistic_regression_model.joblib"
    joblib.dump(model, model_path)
    model_file = JoblibSerializedFile(model_path)
    return model_file


@workflow
def training_workflow(
    logistic_regression: LogisticRegressionInterface = LogisticRegressionInterface(
        max_iter=2000
    ),
    # linear_regression: LinearRegressionInterface = LinearRegressionInterface(),
) -> JoblibSerializedFile:
    """
    Put all of the steps together into a single workflow.
    """
    data = get_data()
    processed_data = process_data(data=data)
    return train_model(
        data=processed_data,
        logistic_regression=logistic_regression,
    )


if __name__ == "__main__":
    # Execute the workflow, simply by invoking it like a function and passing in
    # the necessary parameters
    print(f"Running process_data() { process_data() }")
    print(f"Running training_workflow() { training_workflow() }")

# sample_columns = [
#     "alcohol",
#     "malic_acid",
#     "ash",
#     "alcalinity_of_ash",
#     "magnesium",
#     "total_phenols",
#     "flavanoids",
#     "nonflavanoid_phenols",
#     "proanthocyanins",
#     "color_intensity",
#     "hue",
#     "od280/od315_of_diluted_wines",
#     "proline",
#     "target",
# ]

# sample_data = [
#     [13.0, 1.5, 2.3, 15.0, 110, 2.5, 3.0, 0.3, 1.5, 4.0, 1.0, 3.0, 1000, 0],
#     [14.0, 1.6, 2.4, 16.0, 120, 2.6, 3.1, 0.4, 1.6, 5.0, 1.1, 3.1, 1100, 1],
#     [12.5, 1.4, 2.2, 14.0, 100, 2.4, 2.9, 0.2, 1.4, 3.5, 0.9, 2.9, 900, 2],
# ]



import functools
import logging
import pdb
from inspect import getmembers
from pprint import pprint
from types import FunctionType
from typing import Any, Dict, Tuple

import anndata
import colorlog
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scvelo as scv
import seaborn as sns
import torch
from beartype import beartype
from pytorch_lightning.utilities import rank_zero_only
from scipy.sparse import issparse
from scvi.data import synthetic_iid
from sklearn.decomposition import PCA
from termcolor import colored
from torch.nn.functional import relu


def trace(func):
    def wrapper(*args, **kwargs):
        params = list(args) + [f"{k}={v}" for k, v in kwargs.items()]
        params_str = ",\n    ".join(map(str, params))
        print(f"{func.__name__}(\n    {params_str},\n)")
        return func(*args, **kwargs)

    return wrapper


def inv(x: torch.Tensor) -> torch.Tensor:
    """
    Computes the element-wise reciprocal of a tensor.

    Args:
        x (torch.Tensor): Input tensor.

    Returns:
        torch.Tensor: Tensor with element-wise reciprocal of x.

    Examples:
        >>> import torch
        >>> x = torch.tensor([2., 4., 0.5])
        >>> inv(x)
        tensor([0.5000, 0.2500, 2.0000])
    """
    return x.reciprocal()


def log(x):
    """
    Computes the element-wise natural logarithm of a tensor, while clipping
    its values to avoid numerical instability.

    Args:
        x (torch.Tensor): Input tensor.

    Returns:
        torch.Tensor: Tensor with element-wise natural logarithm of x.

    Examples:
        >>> import torch
        >>> x = torch.tensor([0.0001, 0.5, 0.9999])
        >>> log(x)
        tensor([-9.2103e+00, -6.9315e-01, -1.0002e-04])
    """
    eps = torch.finfo(x.dtype).eps
    return torch.log(x.clamp(eps, 1 - eps))


def velocity_dus_dt(alpha, beta, gamma, tau, x):
    """
    Computes the velocity du/dt and ds/dt.

    Args:
        alpha (torch.Tensor): Alpha parameter.
        beta (torch.Tensor): Beta parameter.
        gamma (torch.Tensor): Gamma parameter.
        tau (torch.Tensor): Time points.
        x (Tuple[torch.Tensor, torch.Tensor]): Tuple containing u and s.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple containing du/dt and ds/dt.

    Examples:
        >>> import torch
        >>> alpha = torch.tensor(0.5)
        >>> beta = torch.tensor(0.4)
        >>> gamma = torch.tensor(0.3)
        >>> tau = torch.tensor(2.0)
        >>> x = (torch.tensor(1.0), torch.tensor(0.5))
        >>> velocity_dus_dt(alpha, beta, gamma, tau, x)
        (tensor(0.1000), tensor(0.2500))
    """
    u, s = x
    du_dt = alpha - beta * u
    ds_dt = beta * u - gamma * s
    return du_dt, ds_dt


def rescale_time(dx_dt, t_start, t_end):
    """
    Converts an ODE to be solved on a batch of different time intervals into an
    equivalent system of ODEs to be solved on [0, 1].

    Args:
        dx_dt (Callable): Function representing the ODE system.
        t_start (torch.Tensor): Start time of the time interval.
        t_end (torch.Tensor): End time of the time interval.

    Returns:
        Callable: Function representing the rescaled ODE system.

    Examples:
        >>> import torch
        >>> def dx_dt(t, x):
        ...     return -x
        >>> t_start = torch.tensor(0.0)
        >>> t_end = torch.tensor(1.0)
        >>> rescaled_dx_dt = rescale_time(dx_dt, t_start, t_end)
        >>> rescaled_dx_dt(torch.tensor(0.5), torch.tensor(2.0))
        tensor(-2.)
    """
    dt = t_end - t_start

    @functools.wraps(dx_dt)
    def dx_ds(s, *args):
        # move any batch dimensions in s to the left of all batch dimensions in dt
        # s_shape = (1,) if len(s.shape) == 0 else s.shape  # XXX unnecessary?
        s = s.reshape(s.shape + (1,) * len(dt.shape))
        t = s * dt + t_start
        xts = dx_dt(t, *args)
        if isinstance(xts, tuple):
            xss = tuple(xt * dt for xt in xts)
        else:
            xss = xts * dt
        return xss

    return dx_ds


def ode_mRNA(tau, u0, s0, alpha, beta, gamma):
    """
    Solves the ODE system for mRNA dynamics.

    Args:
        tau (torch.Tensor): Time points.
        u0 (torch.Tensor): Initial value of u.
        s0 (torch.Tensor): Initial value of s.
        alpha (torch.Tensor): Alpha parameter.
        beta (torch.Tensor): Beta parameter.
        gamma (torch.Tensor): Gamma parameter.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple containing the final values of u and s.
    """
    """
    Examples:
        >>> import torch
        >>> tau = torch.tensor(2.0)
        >>> u0 = torch.tensor(1.0)
        >>> s0 = torch.tensor(0.5)
        >>> alpha = torch.tensor(0.5)
        >>> beta = torch.tensor(0.4)
        >>> gamma = torch.tensor(0.3)
        >>> ode_mRNA(tau, u0, s0, alpha, beta, gamma)
        (tensor(0.6703), tensor(0.4596))
    """
    dx_dt = functools.partial(velocity_dus_dt, alpha, beta, gamma)
    dx_dt = rescale_time(
        dx_dt, torch.tensor(0.0, dtype=tau.dtype, device=tau.device), tau
    )
    grid = torch.linspace(0.0, 1.0, 100, dtype=tau.dtype, device=tau.device)
    x0 = (u0.expand_as(tau), s0.expand_as(tau))
    # xts = odeint_adjoint(dx_dt, x0, grid, adjoint_params=(alpha, beta, gamma, tau))
    xts = odeint(dx_dt, x0, grid)
    uts, sts = xts
    ut, st = uts[-1], sts[-1]
    return ut, st


def mRNA(
    tau: torch.Tensor,
    u0: torch.Tensor,
    s0: torch.Tensor,
    alpha: torch.Tensor,
    beta: torch.Tensor,
    gamma: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Computes the mRNA dynamics given the parameters and initial conditions.

    Args:
        tau (torch.Tensor): Time points.
        u0 (torch.Tensor): Initial value of u.
        s0 (torch.Tensor): Initial value of s.
        alpha (torch.Tensor): Alpha parameter.
        beta (torch.Tensor): Beta parameter.
        gamma (torch.Tensor): Gamma parameter.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple containing the final values of u and s.

    Examples:
        >>> import torch
        >>> tau = torch.tensor(2.0)
        >>> u0 = torch.tensor(1.0)
        >>> s0 = torch.tensor(0.5)
        >>> alpha = torch.tensor(0.5)
        >>> beta = torch.tensor(0.4)
        >>> gamma = torch.tensor(0.3)
        >>> mRNA(tau, u0, s0, alpha, beta, gamma)
        (tensor(1.1377), tensor(0.9269))
    """
    expu, exps = torch.exp(-beta * tau), torch.exp(-gamma * tau)

    # invalid values caused by below codes:
    # gamma equals beta will raise inf, inf * 0 leads to nan
    expus = (alpha - u0 * beta) * inv(gamma - beta) * (exps - expu)
    # solution 1: conditional zero filling
    # solution 1 issue:AutoDelta map_estimate of alpha,beta,gamma,switching will become nan, thus u_inf/s_inf/ut/st all lead to nan
    # expus = torch.where(torch.isclose(gamma, beta), expus.new_zeros(1), expus)

    ut = u0 * expu + alpha / beta * (1 - expu)
    st = (
        s0 * exps + alpha / gamma * (1 - exps) + expus
    )  # remove expus is the most stable, does it theoretically make sense?

    # solution 2: conditional analytical solution
    # solution 2 issue:AutoDelta map_estimate of alpha,beta,gamma,switching will become nan, thus u_inf/s_inf/ut/st all lead to nan
    # On the Mathematics of RNA Velocity I: Theoretical Analysis: Equation (2.12) when gamma == beta
    st2 = (
        s0 * expu + alpha / beta * (1 - expu) - (alpha - beta * u0) * tau * expu
    )
    ##st2 = s0 * expu + alpha / gamma * (1 - expu) - (alpha - gamma * u0) * tau * expu
    st = torch.where(torch.isclose(gamma, beta), st2, st)

    # solution 3: do not use AutoDelta and map_estimate? customize guide function?
    # use solution 3 with st2
    return ut, st


def tau_inv(
    u=None, s=None, u0=None, s0=None, alpha=None, beta=None, gamma=None
):
    """
    Computes the inverse tau given the parameters and initial conditions.

    Args:
        u (torch.Tensor): Value of u.
        s (torch.Tensor): Value of s.
        u0 (torch.Tensor): Initial value of u.
        s0 (torch.Tensor): Initial value of s.
        alpha (torch.Tensor): Alpha parameter.
        beta (torch.Tensor): Beta parameter.
        gamma (torch.Tensor): Gamma parameter.

    Returns:
        torch.Tensor: Inverse tau.

    Examples:
        >>> import torch
        >>> u = torch.tensor(0.6703)
        >>> s = torch.tensor(0.4596)
        >>> u0 = torch.tensor(1.0)
        >>> s0 = torch.tensor(0.5)
        >>> alpha = torch.tensor(0.5)
        >>> beta = torch.tensor(0.4)
        >>> gamma = torch.tensor(0.3)
        >>> tau_inv(u, s, u0, s0, alpha, beta, gamma)
        tensor(3.9736e-07)
    """
    beta_ = beta * inv(gamma - beta)
    xinf = alpha / gamma - beta_ * (alpha / beta)
    tau1 = (
        -1.0 / gamma * log((s - beta_ * u - xinf) * inv(s0 - beta_ * u0 - xinf))
    )

    uinf = alpha / beta
    tau2 = -1 / beta * log((u - uinf) * inv(u0 - uinf))
    tau = torch.where(beta > gamma, tau1, tau2)
    return relu(tau)


def mse_loss_sum(u_model, s_model, u_data, s_data):
    """
    Computes the mean squared error loss sum between the model and data.

    Args:
        u_model (torch.Tensor): Predicted values of u from the model.
        s_model (torch.Tensor): Predicted values of s from the model.
        u_data (torch.Tensor): True values of u from the data.
        s_data (torch.Tensor): True values of s from the data.

    Returns:
        torch.Tensor: Mean squared error loss sum.

    Examples:
        >>> import torch
        >>> u_model = torch.tensor([0.5, 0.6])
        >>> s_model = torch.tensor([0.7, 0.8])
        >>> u_data = torch.tensor([0.4, 0.5])
        >>> s_data = torch.tensor([0.6, 0.7])
        >>> mse_loss_sum(u_model, s_model, u_data, s_data)
        tensor(0.0200)
    """
    return ((u_model - u_data) ** 2 + (s_model - s_data) ** 2).mean(0)


def get_velocity_samples(posterior_samples, model):
    """
    Computes the velocity samples from the posterior samples.

    Args:
        posterior_samples (dict): Dictionary containing posterior samples.
        model: Model used for predictions.

    Returns:
        torch.Tensor: Velocity samples.

    Examples:
        >>> import torch
        >>> posterior_samples = {
        ...     "beta": torch.tensor([[0.4]]),
        ...     "gamma": torch.tensor([[0.3]]),
        ...     "u_scale": torch.tensor([[[1.0, 2.0]]]),
        ...     "s_scale": torch.tensor([[[2.0, 4.0]]]),
        ...     "u": torch.tensor([[1.0, 2.0]]),
        ...     "s": torch.tensor([[0.5, 1.0]])
        ... }
        >>> model = None  # Model is not used in the function
        >>> get_velocity_samples(posterior_samples, model)
        tensor([[0.6500, 1.3000]])
    """
    beta = posterior_samples["beta"].mean(0)[0]
    gamma = posterior_samples["gamma"].mean(0)[0]
    scale = (
        posterior_samples["u_scale"][:, 0, :]
        / posterior_samples["s_scale"][:, 0, :]
    ).mean(0)
    ut = posterior_samples["u"] / scale
    st = posterior_samples["s"]
    v = beta * ut - gamma * st
    return v


def mae(pred_counts, true_counts):
    """
    Computes the mean average error between predicted counts and true counts.

    Args:
        pred_counts (np.ndarray): Predicted counts.
        true_counts (np.ndarray): True counts.

    Returns:
        float: Mean average error.

    Examples:
        >>> import numpy as np
        >>> pred_counts = np.array([[1, 2], [3, 4]])
        >>> true_counts = np.array([[2, 3], [4, 5]])
        >>> mae(pred_counts, true_counts)
        1.0
    """
    error = np.abs(true_counts - pred_counts).sum()
    total = pred_counts.shape[0] * pred_counts.shape[1]
    return (error / total).mean().item()


def init_with_all_cells(
    adata,
    input_type="knn",
    shared_time=True,
    latent_factor="linear",
    latent_factor_size=10,
    plate_size=2,
    num_aux_cells=200,
    init_smooth=True,
):
    ## hard to use unsmoothed data for initialization
    ## always initialize the model with smoothed data
    if "Mu" in adata.layers and "Ms" in adata.layers and init_smooth:
        u_obs = torch.tensor(adata.layers["Mu"], dtype=torch.float32)
        s_obs = torch.tensor(adata.layers["Ms"], dtype=torch.float32)
    elif "spliced" in adata.layers:
        u_obs = torch.tensor(
            adata.layers["unspliced"].toarray()
            if issparse(adata.layers["unspliced"])
            else adata.layers["unspliced"],
            dtype=torch.float32,
        )
        s_obs = torch.tensor(
            adata.layers["spliced"].toarray()
            if issparse(adata.layers["spliced"])
            else adata.layers["spliced"],
            dtype=torch.float32,
        )
    else:
        raise

    ub_u = torch.stack(
        [torch.quantile(u[u > 0], 0.99) for u in torch.unbind(u_obs, axis=1)]
    )
    ub_s = torch.stack(
        [torch.quantile(s[s > 0], 0.99) for s in torch.unbind(s_obs, axis=1)]
    )
    s_mask = (s_obs > 0) & (s_obs <= ub_s)
    u_mask = (u_obs > 0) & (u_obs <= ub_u)
    # include zeros
    training_mask = s_mask & u_mask

    u_scale = torch.stack(
        [u[u > 0].std() for u in torch.unbind(u_obs * training_mask, axis=1)]
    )
    s_scale = torch.stack(
        [s[s > 0].std() for s in torch.unbind(s_obs * training_mask, axis=1)]
    )
    scale = u_scale / s_scale

    lb_steady_u = torch.stack(
        [
            torch.quantile(u[u > 0], 0.98)
            for u in torch.unbind(u_obs * training_mask, axis=1)
        ]
    )
    lb_steady_s = torch.stack(
        [
            torch.quantile(s[s > 0], 0.98)
            for s in torch.unbind(s_obs * training_mask, axis=1)
        ]
    )

    steady_u_mask = training_mask & (u_obs >= lb_steady_u)
    steady_s_mask = training_mask & (s_obs >= lb_steady_s)

    u_obs = u_obs / scale
    u_inf = (u_obs * (steady_u_mask | steady_s_mask)).sum(dim=0) / (
        steady_u_mask | steady_s_mask
    ).sum(dim=0)
    s_inf = (s_obs * steady_s_mask).sum(dim=0) / steady_s_mask.sum(dim=0)

    gamma = (u_obs * steady_s_mask * s_obs).sum(axis=0) / (
        (steady_s_mask * s_obs) ** 2
    ).sum(axis=0) + 1e-6
    gamma = torch.where(gamma < 0.05 / scale, gamma * 1.2, gamma)
    gamma = torch.where(gamma > 1.5 / scale, gamma / 1.2, gamma)
    alpha = gamma * s_inf
    beta = alpha / u_inf

    switching = tau_inv(u_inf, s_inf, 0.0, 0.0, alpha, beta, gamma)
    tau = tau_inv(u_obs, s_obs, 0.0, 0.0, alpha, beta, gamma)
    tau = torch.where(tau >= switching, switching, tau)
    tau_ = tau_inv(u_obs, s_obs, u_inf, s_inf, 0.0, beta, gamma)
    tau_ = torch.where(
        tau_ >= tau_[s_obs > 0].max(dim=0)[0],
        tau_[s_obs > 0].max(dim=0)[0],
        tau_,
    )
    ut, st = mRNA(tau, 0.0, 0.0, alpha, beta, gamma)
    ut_, st_ = mRNA(tau_, u_inf, s_inf, 0.0, beta, gamma)

    u_scale_ = u_scale / scale
    state_on = ((ut - u_obs) / u_scale_) ** 2 + ((st - s_obs) / s_scale) ** 2
    state_off = ((ut_ - u_obs) / u_scale_) ** 2 + ((st_ - s_obs) / s_scale) ** 2
    cell_gene_state_logits = state_on - state_off
    cell_gene_state = cell_gene_state_logits < 0
    t = torch.where(cell_gene_state_logits < 0, tau, tau_ + switching)
    init_values = {
        "alpha": alpha,
        "beta": beta,
        "gamma": gamma,
        "switching": switching,
        "latent_time": t,
        "u_scale": u_scale,
        "s_scale": s_scale,
        "u_inf": u_inf,
        "s_inf": s_inf,
    }
    if input_type == "knn":
        init_values["mask"] = training_mask

    init_values["cell_gene_state"] = cell_gene_state.int()
    if latent_factor == "linear":
        if "spliced" in adata.layers:
            u_obs = torch.tensor(
                adata.layers["unspliced"].toarray()
                if issparse(adata.layers["unspliced"])
                else adata.layers["unspliced"],
                dtype=torch.float32,
            )
            s_obs = torch.tensor(
                adata.layers["spliced"].toarray()
                if issparse(adata.layers["spliced"])
                else adata.layers["spliced"],
                dtype=torch.float32,
            )
        test = np.hstack([u_obs, s_obs])
        pca = PCA(n_components=latent_factor_size)
        pca.fit(test)
        X_train_pca = pca.transform(test)
        init_values["cell_codebook"] = torch.tensor(pca.components_)
        init_values["u_pcs_mean"] = torch.tensor(pca.mean_[: adata.shape[1]])
        init_values["s_pcs_mean"] = torch.tensor(pca.mean_[adata.shape[1] :])
        if plate_size == 2:
            init_values["cell_code"] = (
                torch.tensor(X_train_pca).unsqueeze(-1).transpose(-1, -2)
            )
        else:
            init_values["cell_code"] = torch.tensor(X_train_pca)

    if num_aux_cells > 0:
        np.random.seed(99)
        if "cytotrace" in adata.obs.columns:
            order_aux = np.array_split(
                np.sort(adata.obs["cytotrace"].values), 50
            )
            order_aux_list = []
            for i in order_aux:
                order_aux_list.append(
                    np.random.choice(
                        np.where(
                            (adata.obs["cytotrace"].values > i[0])
                            & (adata.obs["cytotrace"].values < i[-1])
                        )[0]
                    )
                )
            order_aux = np.array(order_aux_list)
        else:
            order_aux = np.argsort(
                (adata.layers["spliced"].toarray() > 0).sum(axis=1)
            )[::-1]

        init_values["order_aux"] = torch.from_numpy(
            order_aux[:num_aux_cells].copy()
        )

        if input_type == "raw":
            u_obs = adata.layers["raw_unspliced"].toarray()
            s_obs = adata.layers["raw_spliced"].toarray()
        init_values["aux_u_obs"] = torch.tensor(
            u_obs[order_aux][:num_aux_cells]
        )
        init_values["aux_s_obs"] = torch.tensor(
            s_obs[order_aux][:num_aux_cells]
        )
        init_values["cell_gene_state_aux"] = cell_gene_state[:num_aux_cells]
        init_values["latent_time_aux"] = t[:num_aux_cells]

        if latent_factor == "linear":
            init_values["cell_code_aux"] = (
                torch.tensor(X_train_pca)
                .unsqueeze(-1)
                .transpose(-1, -2)[:num_aux_cells]
            )

    if shared_time:
        cell_time = t.mean(dim=-1, keepdims=True)
        init_values["cell_time"] = cell_time
        init_values["latent_time"] = init_values["latent_time"] - cell_time

        if num_aux_cells > 0:
            init_values["cell_time_aux"] = cell_time[:num_aux_cells]
            init_values["latent_time_aux"] = (
                init_values["latent_time_aux"] - init_values["cell_time_aux"]
            )

    for key in init_values:
        print(key, init_values[key].shape)
        print(init_values[key].isnan().sum())
        assert init_values[key].isnan().sum() == 0
    return init_values


def ensure_numpy_array(obj):
    return obj.toarray() if hasattr(obj, "toarray") else obj


def mae_evaluate(posterior_samples, adata):
    maes_list = []
    labels = []
    if isinstance(posterior_samples, tuple):
        for model_label, model_obj, split_index in zip(
            ["Poisson train", "Poisson valid"],
            posterior_samples[:2],
            posterior_samples[2:],
        ):
            for sample in range(model_obj["u"].shape[0]):
                maes_list.append(
                    mae(
                        np.hstack(
                            [model_obj["u"][sample], model_obj["s"][sample]]
                        ),
                        np.hstack(
                            [
                                adata.layers["raw_unspliced"].toarray()[
                                    split_index
                                ],
                                adata.layers["raw_spliced"].toarray()[
                                    split_index
                                ],
                            ]
                        ),
                    )
                )
                labels.append(model_label)
    else:
        for model_label, model_obj in zip(
            ["Poisson all cells"], [posterior_samples]
        ):
            for sample in range(model_obj["u"].shape[0]):
                maes_list.append(
                    mae(
                        np.hstack(
                            [model_obj["u"][sample], model_obj["s"][sample]]
                        ),
                        np.hstack(
                            [
                                ensure_numpy_array(
                                    adata.layers["raw_unspliced"]
                                ),
                                ensure_numpy_array(adata.layers["raw_spliced"]),
                            ]
                        ),
                    )
                )
                labels.append(model_label)

    import pandas as pd

    fig, ax = plt.subplots()
    fig.set_size_inches(10, 4)
    df = pd.DataFrame({"MAE": maes_list, "label": labels})
    sns.boxplot(x="label", y="MAE", data=df, ax=ax)
    ax.tick_params(axis="x", rotation=90)
    print(df.groupby("label").mean())
    return df


def debug(x):
    if torch.any(torch.isnan(x)):
        print("nan number: ", torch.isnan(x).sum())
        pdb.set_trace()


def site_is_discrete(site: dict) -> bool:
    return (
        site["type"] == "sample"
        and not site["is_observed"]
        and getattr(site["fn"], "has_enumerate_support", False)
    )


def get_pylogger(name=__name__, log_level="DEBUG") -> logging.Logger:
    """Initializes multi-GPU-friendly python command line logger."""

    formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(levelname)s:%(name)s: %(message)s"
        # "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
        # datefmt=None,
        # reset=True,
        # log_colors={
        #     "debug": "cyan",
        #     "info": "green",
        #     "warning": "yellow",
        #     "error": "red",
        #     "exception": "red",
        #     "fatal": "red",
        #     "critical": "red",
        #     },
    )

    handler = colorlog.StreamHandler()
    handler.setFormatter(formatter)
    logger = colorlog.getLogger(name)
    logger.setLevel(log_level)
    logger.addHandler(handler)
    logger.propagate = False

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    logging_levels = (
        "debug",
        "info",
        "warning",
        "error",
        "exception",
        "fatal",
        "critical",
    )
    for level in logging_levels:
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def attributes(obj):
    """
    get object attributes
    """
    disallowed_names = {
        name
        for name, value in getmembers(type(obj))
        if isinstance(value, FunctionType)
    }
    return {
        name: getattr(obj, name)
        for name in dir(obj)
        if name[0] != "_"
        and name not in disallowed_names
        and hasattr(obj, name)
    }


def print_attributes(obj):
    """
    print object attributes
    """
    pprint(attributes(obj))


@beartype
def pretty_log_dict(d: dict) -> str:
    dict_as_string = ""
    for key, value in d.items():
        key_colored = colored(key, "green")
        value_lines = str(value).split("\n")
        value_colored = "\n".join(
            colored(line, "white") for line in value_lines
        )
        dict_as_string += f"{key_colored}:\n{value_colored}\n"
    return dict_as_string


@beartype
def pretty_print_dict(d: dict):
    print(pretty_log_dict(d))


def filter_startswith_dict(dictionary_with_underscore_keys):
    """Remove entries from a dictionary whose keys start with an underscore.

    Args:
        dictionary_with_underscore_keys (dict): Dictionary to be filtered.

    Returns:
        dict: Filtered dictionary.
    """
    return {
        k: v
        for k, v in dictionary_with_underscore_keys.items()
        if not k.startswith("_")
    }


def print_anndata(anndata_obj):
    """
    Print a formatted representation of an AnnData object.

    This function produces a custom output for the AnnData object with each
    element of obs, var, uns, obsm, varm, layers, obsp, varp indented and
    displayed on a new line.

    Args:
        anndata_obj (anndata.AnnData): The AnnData object to be printed.

    Raises:
        AssertionError: If the input object is not an instance of anndata.AnnData.

    Examples:
        >>> import anndata
        >>> import numpy as np
        >>> import pandas as pd
        >>> np.random.seed(42)
        >>> X = np.random.randn(10, 5)
        >>> obs = pd.DataFrame({"clusters_coarse": np.random.randint(0, 2, 10),
        ...                     "clusters": np.random.randint(0, 2, 10),
        ...                     "S_score": np.random.rand(10),
        ...                     "G2M_score": np.random.rand(10)})
        >>> var = pd.DataFrame({"gene_name": [f"gene_{i}" for i in range(5)]})
        >>> adata = anndata.AnnData(X, obs=obs, var=var)
        >>> print_anndata(adata)  # doctest: +NORMALIZE_WHITESPACE
        AnnData object with n_obs × n_vars = 10 × 5
            obs:
                clusters_coarse,
                clusters,
                S_score,
                G2M_score,
            var:
                gene_name,
    """
    assert isinstance(
        anndata_obj, anndata.AnnData
    ), "Input object must be of type AnnData."

    def format_elements(elements):
        formatted = "\n".join([f"        {elem}," for elem in elements])
        return formatted

    print(
        f"AnnData object with n_obs × n_vars = {anndata_obj.n_obs} × {anndata_obj.n_vars}"
    )

    properties = {
        "obs": anndata_obj.obs.columns,
        "var": anndata_obj.var.columns,
        "uns": anndata_obj.uns.keys(),
        "obsm": anndata_obj.obsm.keys(),
        "varm": anndata_obj.varm.keys(),
        "layers": anndata_obj.layers.keys(),
        "obsp": anndata_obj.obsp.keys(),
        "varp": anndata_obj.varp.keys(),
    }

    for prop_name, elements in properties.items():
        if len(elements) > 0:
            print(f"    {prop_name}:\n{format_elements(elements)}")


def generate_sample_data(
    n_obs: int = 100,
    n_vars: int = 12,
    alpha: float = 5,
    beta: float = 0.5,
    gamma: float = 0.3,
    alpha_: float = 0,
    noise_model: str = "gillespie",
    random_seed: int = 0,
) -> anndata.AnnData:
    """
    Generate synthetic single-cell RNA sequencing data with spliced and unspliced layers.
    If using the "iid" noise model, the data will be generated with scvi.data.synthetic_iid.
    If using the "normal" or "gillespie" noise model, the data will be generated with
    scvelo.datasets.simulation accounting for the given expression dynamics parameters.

    Args:
        n_obs (int, optional): Number of observations (cells). Default is 100.
        n_vars (int, optional): Number of variables (genes). Default is 12.
        alpha (float, optional): Transcription rate. Default is 5.
        beta (float, optional): Splicing rate. Default is 0.5.
        gamma (float, optional): Degradation rate. Default is 0.3.
        alpha_ (float, optional): Additional transcription rate. Default is 0.
        noise_model (str, optional): Noise model to be used. Must be one of 'iid', 'gillespie', or 'normal'. Default is 'gillespie'.
        random_seed (int, optional): Random seed for reproducibility. Default is 0.

    Returns:
        anndata.AnnData: An AnnData object containing the generated synthetic data.

    Raises:
        ValueError: If noise_model is not one of 'iid', 'gillespie', or 'normal'.

    Examples:
        >>> from pyrovelocity.utils import generate_sample_data, print_anndata
        >>> adata = generate_sample_data(random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(n_obs=50, n_vars=10, noise_model="normal", random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(n_obs=50, n_vars=10, noise_model="iid", random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(noise_model="wishful thinking")
        Traceback (most recent call last):
            ...
        ValueError: noise_model must be one of 'iid', 'gillespie', 'normal'
    """
    if noise_model == "iid":
        adata = synthetic_iid(
            batch_size=n_obs,
            n_genes=n_vars,
            n_batches=1,
            n_labels=1,
        )
        adata.layers["spliced"] = adata.X.copy()
        adata.layers["unspliced"] = adata.X.copy()
    elif noise_model in {"gillespie", "normal"}:
        adata = scv.datasets.simulation(
            random_seed=random_seed,
            n_obs=n_obs,
            n_vars=n_vars,
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            alpha_=alpha_,
            noise_model=noise_model,
        )
    else:
        raise ValueError(
            "noise_model must be one of 'iid', 'gillespie', 'normal'"
        )
    return adata


def anndata_counts_to_df(adata):
    spliced_df = pd.DataFrame(
        ensure_numpy_array(adata.layers["raw_spliced"]),
        index=list(adata.obs_names),
        columns=list(adata.var_names),
    )
    unspliced_df = pd.DataFrame(
        ensure_numpy_array(adata.layers["raw_unspliced"]),
        index=list(adata.obs_names),
        columns=list(adata.var_names),
    )

    spliced_melted = spliced_df.reset_index().melt(
        id_vars="index", var_name="var_name", value_name="spliced"
    )
    unspliced_melted = unspliced_df.reset_index().melt(
        id_vars="index", var_name="var_name", value_name="unspliced"
    )

    df = spliced_melted.merge(unspliced_melted, on=["index", "var_name"])

    df = df.rename(columns={"index": "obs_name"})

    total_obs = adata.n_obs
    total_var = adata.n_vars

    max_spliced = adata.layers["raw_spliced"].max()
    max_unspliced = adata.layers["raw_unspliced"].max()

    return (
        df,
        total_obs,
        total_var,
        max_spliced,
        max_unspliced,
    )



import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics.pairwise import euclidean_distances, rbf_kernel
from sklearn.neighbors import KernelDensity


class VMF:
    """
    A class for creating and handling von Mises-Fisher distributions in N dimensions.

    Attributes:
        mu (array-like): The mean direction of the distribution.
        kappa (float): The concentration parameter of the distribution.
        dim (int): The number of dimensions.
    """

    def __init__(self, mu, kappa):
        """
        Stores parameters for a VMF object.

        Args:
            mu (array-like): The mean direction of the distribution.
            kappa (float): The concentration parameter of the distribution.
        """

        self.mu = mu
        self.kappa = kappa
        self.dim = len(mu)

    def sample(self, num_samples):
        """
        Samples from the von Mises-Fisher distribution.

        Args:
            num_samples (int): The number of samples to draw.

        Returns:
            ndarray: The sampled points in the form of a num_samples by dim array.
        """
        result = np.zeros((num_samples, self.dim))

        for i in range(num_samples):
            w = self._sample_weight()
            v = self._sample_orthogonal_to()
            result[i, :] = np.sqrt(1 - w**2) * v + w * self.mu

        return result

    def _sample_weight(self):
        """
        Samples the weight component used for generating N-D vMF samples.

        Returns:
            float: The sampled weight.
        """
        b = (
            -2 * self.kappa + np.sqrt(4 * self.kappa**2 + (self.dim - 1) ** 2)
        ) / (self.dim - 1)
        x = (1 - b) / (1 + b)
        c = self.kappa * x + (self.dim - 1) * np.log(1 - x**2)

        while True:
            z = np.random.beta(0.5 * (self.dim - 1), 0.5 * (self.dim - 1))
            w = (1 - (1 + b) * z) / (1 - (1 - b) * z)
            u = np.random.uniform(0, 1)
            if self.kappa * w + (self.dim - 1) * np.log(
                1 - x * w
            ) - c >= np.log(u):
                return w

    def _sample_orthogonal_to(self):
        """
        Sample a vector orthogonal to the mean direction.

        Returns:
            ndarray: The sampled orthogonal vector.
        """
        v = np.random.randn(self.dim)
        proj_mu_v = self.mu * np.dot(self.mu, v)
        return (v - proj_mu_v) / np.linalg.norm(v - proj_mu_v)


def sample_uniform(dimension, sample_number):
    """
    Samples from a uniform distribution on the N-D unit hypersphere.

    Args:
        sample_number (int): The number of samples to draw.

    Returns:
        ndarray: The sampled points in the form of a sample_number by dimension array.
    """
    result = np.random.randn(sample_number, dimension)
    result /= np.linalg.norm(result, axis=1)[:, np.newaxis]
    return result


def kl_divergence(p, q, bandwidth=None):
    """
    Computes the Kullback-Leibler divergence between two distributions, estimated using kernel density estimation.
    This uses Scott's rule to estimate the bandwidth parameter.

    Scott, D.W. (1992) Multivariate Density Estimation. Theory, Practice and Visualization. New York: Wiley.

    Args:
        p (ndarray): Samples from the first distribution.
        q (ndarray): Samples from the second distribution.
        bandwidth (float, optional): The bandwidth parameter for the kernel density estimation.

    Returns:
        float: The Kullback-Leibler divergence between the first and second distributions.
        float: The bandwidth parameter for the kernel density estimation used in the Kullback-Leibler divergence computation.
    """

    sample_number, dimension = p.shape

    if bandwidth is None:
        bandwidth = sample_number ** (-1 / (dimension + 4))

    p_kde = KernelDensity(kernel="gaussian", bandwidth=bandwidth).fit(p)
    q_kde = KernelDensity(kernel="gaussian", bandwidth=bandwidth).fit(q)

    p_scores = p_kde.score_samples(p)
    q_scores = q_kde.score_samples(p)

    kl_div = np.sum(p_scores - q_scores) / sample_number

    return kl_div, bandwidth


def js_divergence(p, q, bandwidth=None):
    """
    Computes the Jensen-Shannon divergence between two distributions using `kl_divergence`.

    Args:
        p (ndarray): Samples from the first distribution.
        q (ndarray): Samples from the second distribution.
        bandwidth (float, optional): The bandwidth parameter for the kernel density estimation.

    Returns:
        float: The Jensen-Shannon divergence between the first and second distributions.
    """

    # Calculate the KL divergences and ignore bandwidths
    kl_pq, bandwidth = kl_divergence(p, q, bandwidth)
    kl_qp, _ = kl_divergence(q, p, bandwidth)

    # Calculate the JS divergence
    js_div = 0.5 * kl_pq + 0.5 * kl_qp

    return js_div, bandwidth


def mmd(p, q, gamma=None):
    """
    Compute the Maximum Mean Discrepancy (MMD) between two multidimensional distributions using the RBF kernel.

    Args:
        p (ndarray): Samples from the first distribution.
        q (ndarray): Samples from the second distribution.
        gamma (float, optional): The gamma parameter for the RBF kernel. If not provided, it will be set to the inverse
                                 of the median of the pairwise distances in the combined data.

    Returns:
        float: The Maximum Mean Discrepancy between the first and second distributions.
    """
    combined = np.vstack((p, q))

    if gamma is None:
        pairwise_dists = euclidean_distances(combined, combined)

        gamma = 1.0 / np.median(pairwise_dists)

    p_kernel = rbf_kernel(p, p, gamma)
    q_kernel = rbf_kernel(q, q, gamma)
    cross_kernel = rbf_kernel(p, q, gamma)

    p_mean = np.mean(p_kernel)
    q_mean = np.mean(q_kernel)
    cross_mean = np.mean(cross_kernel)

    mmd = p_mean - 2 * cross_mean + q_mean

    return mmd, gamma


def plot_samples_and_kde(
    ax, samples_vmf, samples_uniform, kappa, dim=3, bandwidth=0.1
):
    if dim not in [2, 3]:
        raise ValueError("Dimension must be 2 or 3.")

    if dim == 3:
        ax.scatter(
            samples_vmf[:, 0],
            samples_vmf[:, 1],
            samples_vmf[:, 2],
            color="darkgreen",
        )
        ax.set_title(rf"VMF distribution ($\kappa$={kappa})")
        ax.set_xlim([-1, 1])
        ax.set_ylim([-1, 1])
        ax.set_zlim([-1, 1])

    else:
        kde_vmf = KernelDensity(bandwidth=bandwidth).fit(samples_vmf)
        kde_uniform = KernelDensity(bandwidth=bandwidth).fit(samples_uniform)

        # grid
        num_points = 100
        x = np.linspace(
            min(samples_vmf[:, 0]), max(samples_vmf[:, 0]), num_points
        )
        y = np.linspace(
            min(samples_vmf[:, 1]), max(samples_vmf[:, 1]), num_points
        )
        X, Y = np.meshgrid(x, y)
        xy = np.vstack([X.ravel(), Y.ravel()]).T

        # evaluate
        Z_vmf = np.exp(kde_vmf.score_samples(xy)).reshape(X.shape)
        Z_uniform = np.exp(kde_uniform.score_samples(xy)).reshape(X.shape)

        # 2-D vMF distribution with contours
        ax.scatter(samples_vmf[:, 0], samples_vmf[:, 1], color="darkgreen")
        ax.contour(X, Y, Z_vmf, cmap="viridis")
        ax.set_title(rf"VMF distribution ($\kappa$={kappa})")
        ax.set_xlim([-1.3, 1.3])
        ax.set_ylim([-1.3, 1.3])


def sample_and_plot_vMF(num_samples, kappas, dim):
    """
    Demonstrates sampling from an N-dimensional von Mises-Fisher
    distribution with different scale parameters, and compares them to a
    uniform distribution on the unit hypersphere using the Kullback-Leibler divergence.
    This quantifies the extent to which samples of N-dimensional vectors are
    pointing in a coherent direction versus uniformly random directions.

    Args:
        num_samples (int): The number of samples to draw from each distribution.
        kappas (list of float): The scale parameters for the von Mises-Fisher distributions.
        dim (int): The number of dimensions for the distributions.
    """

    mu = np.ones(dim) / np.sqrt(dim)
    print(mu)
    vmf = VMF(mu, kappas[0])
    uniform_samples = sample_uniform(dim, num_samples)

    kl_divs = []

    if dim in [2, 3]:
        fig = plt.figure(figsize=(18, 18))

    # Sample from von Mises-Fisher distributions and
    # compute KL divergence from uniform samples
    for idx, kappa in enumerate(kappas):
        vmf.kappa = kappa  # Update kappa
        vmf_samples = vmf.sample(num_samples)
        kl_div, bandwidth = mmd(vmf_samples, uniform_samples)
        print(f"KL divergence for kappa = {kappa}: {kl_div:.3f}")

        kl_divs.append(kl_div)

        if dim == 2:
            ax = fig.add_subplot(4, 4, idx + 1)
            plot_samples_and_kde(
                ax, vmf_samples, uniform_samples, kappa, dim, bandwidth
            )
        elif dim == 3:
            ax = fig.add_subplot(4, 4, idx + 1, projection="3d")
            plot_samples_and_kde(
                ax, vmf_samples, uniform_samples, kappa, dim, bandwidth
            )

    if dim in [2, 3]:
        plt.subplots_adjust(wspace=0.2, hspace=0.2)

        plt.savefig(f"vmf_samples_{dim}D_kde_subplots.pdf", format="pdf")
        plt.savefig(f"vmf_samples_{dim}D_kde_subplots.png", format="png")
        plt.show()

    # Plot KL divergence vs scale parameter
    plt.figure(figsize=(6, 6))
    plt.plot(kappas, kl_divs, "o", color="darkgreen", markersize=10)
    plt.xlabel(r"$\kappa$")
    plt.ylabel("KL Divergence")
    plt.title(f"{dim}D KL Divergence vs Scale parameter of vMF distribution")
    plt.savefig(f"kl_div_vs_kappa_{dim}D.pdf", format="pdf")
    plt.savefig(f"kl_div_vs_kappa_{dim}D.png", format="png")
    plt.show()

    return bandwidth


if __name__ == "__main__":
    """
    Compute examples over a range of dimensions and scale parameters.

    > montage -geometry 800x kl_div_vs_kappa_2D.png kl_div_vs_kappa_3D.png \
        kl_div_vs_kappa_4D.png kl_div_vs_kappa_10D.png kl_div_vs_kappa_100D.png \
        kl_div_vs_kappa_1000D.png ND_KLD_vs_vMF_kappa.png
    """
    num_samples = 30
    kappas = [1, 3, 4, 6, 8, 10, 15, 20, 35, 50, 75, 100, 150, 200, 250, 300]
    dims = [2, 3, 4, 10, 20, 40, 80, 100, 1000]

    for dim in dims:
        print(f"\n==Dimension: {dim}==\n")
        bandwidth = sample_and_plot_vMF(num_samples, kappas, dim)
        print(f"KD bandwidth: {bandwidth}")



import os
from pathlib import Path
from typing import List, Optional, Tuple

import anndata
import anndata._core.anndata
import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import scvelo as scv
from beartype import beartype
from scipy.sparse import issparse

import pyrovelocity.datasets
from pyrovelocity.cytotrace import cytotrace_sparse
from pyrovelocity.logging import configure_logging
from pyrovelocity.utils import (
    ensure_numpy_array,
    print_anndata,
    print_attributes,
)

logger = configure_logging(__name__)


@beartype
def preprocess_dataset(
    data_path: str | Path = "data/external/simulated.h5ad",
    process_args: dict = {"count_threshold": 0},
    data_processed_path: str | Path = "data/processed",
) -> Path:
    """
    Preprocess dataset.

    Args:
        data_path (str, optional): _description_. Defaults to "data/external/simulated.h5ad".
        process_args (_type_, optional): _description_. Defaults to {"count_thres": 0}.
        data_processed_path (str, optional): _description_. Defaults to "data/processed".

    Returns:
        Path: path to processed dataset.

    Examples:
        >>> from pyrovelocity.data import download_dataset, subset
        >>> from pyrovelocity.preprocess import preprocess_dataset
        >>> pancreas_path = download_dataset(
        ...     data_set_name="pancreas",
        ... ) # xdoctest: +SKIP
        >>> preprocess_dataset(
        ...     data_path=pancreas_path,
        ... ) # xdoctest: +SKIP
        >>> pbmc68k_path = download_dataset(
        ...     data_set_name="pbmc68k",
        ... ) # xdoctest: +SKIP
        >>> _, subset_output_path = subset(
        ...    file_path=pbmc68k_path,
        ...    n_obs=1000,
        ...    save_subset=True,
        ... ) # xdoctest: +SKIP
        >>> preprocessed_path = preprocess_dataset(
        ...     data_path=subset_output_path
        ... ) # xdoctest: +SKIP
    """
    data_set_name = Path(data_path).stem
    processed_path = os.path.join(
        data_processed_path, f"{data_set_name}_processed.h5ad"
    )
    thresh_histogram_path = os.path.join(
        data_processed_path, f"{data_set_name}_thresh_histogram.pdf"
    )
    logger.info(
        f"\n\nVerifying existence of path for:\n\n"
        f"  processed data: {data_processed_path}\n"
    )
    Path(data_processed_path).mkdir(parents=True, exist_ok=True)

    logger.info(
        f"\n\nPreprocessing {data_set_name} data :\n\n"
        f"  from external: {data_path}\n"
        f"  to processed: {processed_path}\n"
    )

    if os.path.isfile(processed_path) and os.access(processed_path, os.R_OK):
        logger.info(f"{processed_path} exists")
        return Path(processed_path)
    else:
        logger.info(f"generating {processed_path} ...")
        adata, output_processed_path = preprocess_data(
            data_set_name=data_set_name,
            data_path=data_path,
            processed_path=processed_path,
            thresh_histogram_path=thresh_histogram_path,
            **process_args,
        )
        print_attributes(adata)

        if os.path.isfile(processed_path) and os.access(
            processed_path, os.R_OK
        ):
            logger.info(f"successfully generated {processed_path}")
            return Path(processed_path)
        else:
            logger.error(f"cannot find and read {processed_path}")


@beartype
def preprocess_data(
    data_set_name: str,
    adata: Optional[anndata._core.anndata.AnnData] = None,
    data_path: Optional[str | Path] = None,
    n_top_genes: int = 2000,
    min_shared_counts: int = 30,
    processed_path: Optional[str] = None,
    force: bool = False,
    process_cytotrace: bool = False,
    use_maximum_likelihood_subset: bool = False,
    count_threshold: int = 0,
    n_pcs: int = 30,
    n_neighbors: int = 30,
    thresh_histogram_path: Optional[str] = None,
) -> Tuple[anndata._core.anndata.AnnData, str | Path]:
    """
    Preprocess data.

    Args:
        data_set_name (str): name of data set.
        adata (AnnData, optional): AnnData object containing the data set to preprocess.
        data_path (str | Path, optional): path to data set. Defaults to None.
        n_top_genes (int, optional): number of genes to retain. Defaults to 2000.
        min_shared_counts (int, optional): minimum shared counts. Defaults to 30.
        processed_path (str, optional): path to read/write processed AnnData. Defaults to None.

    Returns:
        anndata._core.anndata.AnnData: processed AnnData object
    """
    if processed_path is None:
        processed_path = f"{data_set_name}_processed.h5ad"
    if (
        os.path.isfile(processed_path)
        and os.access(processed_path, os.R_OK)
        and (not force)
    ):
        return sc.read(
            filename=processed_path, sparse=True, cache=True
        ), processed_path
    elif adata is not None:
        pass
    elif data_path is not None:
        adata = sc.read(filename=data_path, sparse=True, cache=True)

    if "raw_unspliced" not in adata.layers:
        copy_raw_counts(adata)
        print_anndata(adata)

    if process_cytotrace:
        print("Processing data with cytotrace ...")
        cytotrace_sparse(adata, layer="spliced")

    scv.pp.filter_and_normalize(
        adata, min_shared_counts=min_shared_counts, n_top_genes=n_top_genes
    )
    plot_high_us_genes(
        adata=adata,
        thresh_histogram_path=thresh_histogram_path,
        minlim_u=count_threshold,
        minlim_s=count_threshold,
        unspliced_layer="raw_unspliced",
        spliced_layer="raw_spliced",
    )
    adata = get_high_us_genes(
        adata,
        minlim_u=count_threshold,
        minlim_s=count_threshold,
        unspliced_layer="raw_unspliced",
        spliced_layer="raw_spliced",
    )
    scv.pp.moments(adata, n_pcs=n_pcs, n_neighbors=n_neighbors)
    scv.tl.recover_dynamics(adata, n_jobs=-1, use_raw=False)

    if "pbmc68k" in data_set_name:
        scv.tl.velocity(adata, mode="stochastic", use_raw=False)
    else:
        scv.tl.velocity(adata, mode="dynamical", use_raw=False)

    if ("X_umap" not in adata.obsm.keys()) or (data_set_name == "larry_tips"):
        scv.tl.umap(adata)
    if "leiden" not in adata.obs.keys():
        sc.tl.leiden(adata)
    if use_maximum_likelihood_subset:
        top_genes = (
            adata.var["fit_likelihood"].sort_values(ascending=False).index
        )
        print(top_genes[:10])
        adata = adata[:, top_genes[:3]].copy()
    scv.tl.velocity_graph(adata, n_jobs=-1)

    if data_set_name in [
        "larry",
        "larry_mono",
        "larry_neu",
        "larry_multilineage",
    ]:
        scv.tl.velocity_embedding(adata, basis="emb")
    else:
        scv.tl.velocity_embedding(adata)

    scv.tl.latent_time(adata)

    print_anndata(adata)
    adata.write(processed_path)

    return adata, processed_path


def copy_raw_counts(
    adata: anndata._core.anndata.AnnData,
) -> anndata._core.anndata.AnnData:
    """
    Copy unspliced and spliced raw counts to adata.layers and adata.obs.

    Args:
        adata (anndata._core.anndata.AnnData): AnnData object

    Returns:
        anndata._core.anndata.AnnData: AnnData object with raw counts.

    Examples:
        >>> from pyrovelocity.utils import generate_sample_data
        >>> adata = generate_sample_data()
        >>> copy_raw_counts(adata)
    """
    adata.layers["raw_unspliced"] = adata.layers["unspliced"]
    print("'raw_unspliced', raw unspliced counts (adata.layers)")
    adata.layers["raw_spliced"] = adata.layers["spliced"]
    print("'raw_spliced', raw spliced counts (adata.layers)")
    adata.obs["u_lib_size_raw"] = (
        adata.layers["raw_unspliced"].toarray().sum(-1)
        if issparse(adata.layers["raw_unspliced"])
        else adata.layers["raw_unspliced"].sum(-1)
    )
    print("'u_lib_size_raw', unspliced library size (adata.obs)")
    adata.obs["s_lib_size_raw"] = (
        adata.layers["raw_spliced"].toarray().sum(-1)
        if issparse(adata.layers["raw_spliced"])
        else adata.layers["raw_spliced"].sum(-1)
    )
    print("'s_lib_size_raw', spliced library size (adata.obs)")
    return adata


def assign_colors(
    max_spliced: int, max_unspliced: int, minlim_s: int, minlim_u: int
) -> List[str]:
    return [
        "black"
        if (spliced >= minlim_s) & (unspliced >= minlim_u)
        else "lightgrey"
        for spliced, unspliced in zip(max_spliced, max_unspliced)
    ]


def get_thresh_histogram_title_from_path(path):
    title = os.path.basename(path)
    title = os.path.splitext(title)[0]
    title = title.replace("_thresh_histogram", "")
    return title.replace("_", " ")


def plot_high_us_genes(
    adata: anndata.AnnData,
    thresh_histogram_path: str,
    minlim_u: int = 3,
    minlim_s: int = 3,
    unspliced_layer: str = "unspliced",
    spliced_layer: str = "spliced",
) -> Optional[matplotlib.figure.Figure]:
    if (
        adata is None
        or unspliced_layer not in adata.layers
        or spliced_layer not in adata.layers
    ):
        raise ValueError(
            "Invalid data set. Please ensure that adata is an AnnData object"
            "and that the layers 'unspliced' and 'spliced' are present."
        )

    max_unspliced = np.array(
        np.max(ensure_numpy_array(adata.layers[unspliced_layer]), axis=0)
    ).flatten()
    max_spliced = np.array(
        np.max(ensure_numpy_array(adata.layers[spliced_layer]), axis=0)
    ).flatten()

    # create figure
    x = max_spliced
    y = max_unspliced

    title = get_thresh_histogram_title_from_path(thresh_histogram_path)

    colors = assign_colors(max_spliced, max_unspliced, minlim_s, minlim_u)

    left, width = 0.1, 0.65
    bottom, height = 0.1, 0.65
    spacing = 0.005

    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom + height + spacing, width, 0.2]
    rect_histy = [left + width + spacing, bottom, 0.2, height]

    fig = plt.figure(figsize=(4, 4))
    fig.suptitle(title, y=1.00, fontsize=12)

    ax = fig.add_axes(rect_scatter)
    ax_histx = fig.add_axes(rect_histx, sharex=ax)
    ax_histy = fig.add_axes(rect_histy, sharey=ax)
    ax_histx.tick_params(axis="x", labelbottom=False)
    ax_histy.tick_params(axis="y", labelleft=False)

    # scatter plot
    ax.scatter(max_spliced, max_unspliced, s=1, c=colors)
    ax.set_yscale("log")
    ax.set_xscale("log")
    ax.set_xlabel("max. spliced counts")
    ax.set_ylabel("max. unspliced counts")
    if minlim_s > 0:
        ax.axhline(y=minlim_s - 1, color="r", linestyle="--")
    if minlim_u > 0:
        ax.axvline(x=minlim_u - 1, color="r", linestyle="--")

    # histograms
    bins = 50
    hist, bins = np.histogram(x, bins=bins)
    logbins = np.logspace(0, np.log10(bins[-1]), len(bins))
    ax_histx.hist(x, bins=logbins)
    ax_histy.hist(y, bins=logbins, orientation="horizontal")

    for ext in ["", ".png"]:
        fig.savefig(
            f"{thresh_histogram_path}{ext}",
            facecolor=fig.get_facecolor(),
            bbox_inches="tight",
            edgecolor="none",
            dpi=300,
        )

    return fig


def get_high_us_genes(
    adata,
    minlim_u=0,
    minlim_s=0,
    unspliced_layer="unspliced",
    spliced_layer="spliced",
):
    """
    Function to select genes that have spliced and unspliced counts above a
    certain threshold. Genes of which the maximum u and s count is above a set
    threshold are selected. Threshold varies per dataset and influences the
    numbers of genes that are selected.

    Parameters
    ----------
    adata
        Annotated data matrix
    minlim_u: `int` (default: 3)
        Threshold above which the maximum unspliced counts of a gene should fall to be included in the
        list of high US genes.
    minlim_s: `int` (default: 3)
        Threshold above which the maximum spliced counts of a gene should fall to be included in the
        list of high US genes.
    unspliced_layer: `str` (default: 'unspliced')
        Name of layer that contains the unspliced counts.
    spliced_layer: `str` (default: 'spliced')
        Name of layer that contains the spliced counts.
    """
    print("adata.shape before filtering:", adata.shape)
    from scipy import sparse

    # test if layers are not sparse but dense
    for layer in [unspliced_layer, spliced_layer]:
        if sparse.issparse(adata.layers[layer]):
            adata.layers[layer] = adata.layers[layer].todense()

    # get high US genes
    u_genes = np.max(adata.layers[unspliced_layer], axis=0) >= minlim_u
    s_genes = np.max(adata.layers[spliced_layer], axis=0) >= minlim_s
    us_genes = adata.var_names[np.array(u_genes & s_genes).flatten()].values
    adata = adata[:, us_genes]
    for layer in [unspliced_layer, spliced_layer]:
        adata.layers[layer] = sparse.csr_matrix(adata.layers[layer])
    print("adata.shape after filtering:", adata.shape)
    return adata


# TODO: remove deprecated function
def load_pbmc68k(
    data: str = "pbmc68k",
    count_threshold: int = 0,
    thresh_histogram_path: str = None,
) -> anndata._core.anndata.AnnData:
    if data == "pbmc68k":
        adata = pyrovelocity.datasets.pbmc68k()
    elif os.path.isfile(data) and os.access(data, os.R_OK):
        adata = sc.read(data)

    print_anndata(adata)
    if "raw_unspliced" not in adata.layers:
        copy_raw_counts(adata)
        print_anndata(adata)
    # Integrated into pyrovelocity.datasets.pbmc68k()
    # print("Removing duplicate cells and tSNE x-parity in pbmc68k data...")
    # scv.pp.remove_duplicate_cells(adata)
    # adata.obsm["X_tsne"][:, 0] *= -1
    scv.pp.filter_and_normalize(adata, min_shared_counts=30, n_top_genes=2000)
    plot_high_us_genes(
        adata,
        thresh_histogram_path,
        minlim_u=count_threshold,
        minlim_s=count_threshold,
        unspliced_layer="raw_unspliced",
        spliced_layer="raw_spliced",
    )
    adata = get_high_us_genes(
        adata,
        minlim_u=count_threshold,
        minlim_s=count_threshold,
        unspliced_layer="raw_unspliced",
        spliced_layer="raw_spliced",
    )
    scv.pp.moments(adata)
    scv.tl.velocity(adata, mode="stochastic")
    scv.tl.recover_dynamics(adata, n_jobs=-1)

    return adata


# TODO: remove deprecated function
def load_data(
    data: str = "pancreas",
    processed_path: str = None,
    count_threshold: int = 0,
    thresh_histogram_path: str = None,
) -> anndata._core.anndata.AnnData:
    """
    Preprocess data from scvelo.

    Args:
        data (str, optional): data set name. Defaults to scvelo's "pancreas" data set.
        top_n (int, optional): number of genes to retain. Defaults to 2000.
        min_shared_counts (int, optional): minimum shared counts. Defaults to 30.
        eps (float, optional): tolerance. Defaults to 1e-6.
        force (bool, optional): force reprocessing. Defaults to False.
        processed_path (str, optional): path to read/write processed AnnData. Defaults to None.

    Returns:
        anndata._core.anndata.AnnData: processed AnnData object
    """
    print("Dataset name:", data)
    if data == "pancreas":
        adata = pyrovelocity.datasets.pancreas()
    elif data == "bonemarrow":
        adata = pyrovelocity.datasets.bonemarrow()
    elif data == "pbmc68k":
        adata = pyrovelocity.datasets.pbmc68k()
    elif data == "pons":
        adata = pyrovelocity.datasets.pons()
    elif data == "pbmc5k":
        adata = pyrovelocity.datasets.pbmc5k()
    elif data == "pbmc10k":
        adata = pyrovelocity.datasets.pbmc10k()
    elif data == "larry":
        adata = pyrovelocity.datasets.larry()
    elif data == "larry_mono":
        adata = pyrovelocity.datasets.larry_mono()
    elif data == "larry_neu":
        adata = pyrovelocity.datasets.larry_neu()
    elif data == "larry_multilineage":
        adata = pyrovelocity.datasets.larry_multilineage()
    elif data == "larry_tips":
        adata = pyrovelocity.datasets.larry_tips()
    elif data == "larry_cospar":
        adata = pyrovelocity.datasets.larry_cospar()
    elif data == "larry_cytotrace":
        adata = pyrovelocity.datasets.larry_cytotrace()
    elif data == "larry_dynamical":
        adata = pyrovelocity.datasets.larry_dynamical()
    print_anndata(adata)
    return adata



from typing import Iterable
from typing import Optional

import pyro
import torch
from pyro import poutine
from pyro.contrib import easyguide
from pyro.distributions import Bernoulli
from pyro.distributions import Categorical
from pyro.distributions import Delta
from pyro.distributions import Normal
from pyro.distributions.constraints import positive
from pyro.infer import autoguide
from pyro.infer.autoguide.guides import AutoGuideList
from pyro.nn import PyroModule
from pyro.nn import PyroParam
from scvi.nn import Encoder
from scvi.nn import FCLayers
from torch import nn
from torch.nn.functional import softmax
from torch.nn.functional import softplus

# from ._velocity_model import LatentFactor
from pyrovelocity._velocity_model_deprecated import VelocityModel
from pyrovelocity.utils import mRNA
from pyrovelocity.utils import site_is_discrete
from pyrovelocity.utils import tau_inv


class VelocityGuide(easyguide.EasyGuide):
    def __init__(
        self,
        model: VelocityModel,
        likelihood: str,
        shared_time: bool,
        t_scale: bool,
        latent_factor: str = "none",
        latent_factor_operation: str = "selection",
        model_type: str = "none",
        plate_size: int = 2,
        only_cell_times: bool = False,
        add_offset: bool = False,
        **initial_values,
    ):
        super().__init__(model)
        self.num_genes = model.num_genes
        self.num_cells = model.num_cells
        self.num_aux_cells = model.num_aux_cells

        self.zero = model.zero
        self.one = model.one
        self.shared_time = shared_time
        self.t_scale_on = t_scale
        self.latent_factor = latent_factor
        self.latent_factor_operation = latent_factor_operation
        self.model_type = model_type
        self.only_cell_times = only_cell_times

        self.plate_size = plate_size
        self.event_dim = 0 if self.model_type in ["velocity", "traj"] else 1
        self.likelihood = likelihood
        self.add_offset = add_offset

        self.mask = initial_values.get(
            "mask", torch.ones(self.num_cells, self.num_genes).bool()
        )
        for key in initial_values:
            self.register_buffer(f"{key}_init", initial_values[key])

        if hasattr(self, "alpha_init") and hasattr(
            self, "u_inf_init"
        ):  # with initialization of steady-state
            self.u_inf = PyroParam(
                self.u_inf_init.clone(),
                constraint=positive,
                event_dim=self.event_dim,
            )
            self.s_inf = PyroParam(
                self.s_inf_init.clone(),
                constraint=positive,
                event_dim=self.event_dim,
            )
        else:
            self.u_inf = PyroParam(
                torch.ones(self.num_genes), constraint=positive, event_dim=0
            )
            self.s_inf = PyroParam(
                torch.ones(self.num_genes), constraint=positive, event_dim=0
            )
        ###if not self.shared_time:
        if not (hasattr(self, "alpha_init") and hasattr(self, "u_inf_init")):
            self.cell_time_init = None
            self.cell_time_aux_init = None

        if self.latent_factor != "linear":
            self.cell_code_init = None
            self.cell_code_aux_init = None

        if self.likelihood != "Normal":
            hidden_size = 128
            self.neuralnetwork = nn.Sequential(
                nn.Linear(self.num_genes * 2, hidden_size),
                nn.ReLU(),
                nn.Linear(hidden_size, hidden_size),
                nn.ReLU(),
                nn.Linear(hidden_size, self.num_genes * 2),
            )

        # self.kinetics_encoder = KineticsParamEncoder(10, self.num_cells)
        # self.genetime_encoder = GeneTimeEncoder(10)

    def init(self, site):
        if hasattr(self, "alpha_init") and hasattr(self, "u_inf_init"):
            if self.num_aux_cells > 0:
                if site["name"] == "aux/cell_code":
                    return autoguide.init_to_value(
                        site, values={site["name"]: self.cell_code_aux_init}
                    )
                if site["name"] == "aux/latent_time":
                    return autoguide.init_to_value(
                        site,
                        values={
                            site["name"]: pyro.subsample(
                                self.latent_time_aux_init, event_dim=0
                            )
                        },
                    )
                if site["name"] == "aux/cell_gene_state":
                    return autoguide.init_to_value(
                        site,
                        values={
                            site["name"]: pyro.subsample(
                                self.cell_gene_state_aux_init, event_dim=0
                            )
                        },
                    )
                if site["name"] == "aux/cell_time":
                    return autoguide.init_to_value(
                        site,
                        values={
                            site["name"]: pyro.subsample(
                                self.cell_time_aux_init, event_dim=0
                            )
                        },
                    )
            if site["name"] == "cell_code":
                return autoguide.init_to_value(
                    site,
                    values={
                        site["name"]: pyro.subsample(
                            self.cell_code_init, event_dim=1
                        )
                    },
                )
            if site["name"] == "cell_codebook":
                return autoguide.init_to_value(
                    site,
                    values={
                        site["name"]: pyro.subsample(
                            self.cell_codebook_init, event_dim=1
                        )
                    },
                )
            if site["name"] == "latent_time":
                if self.num_aux_cells > 0:
                    return autoguide.init_to_value(
                        site,
                        values={
                            site["name"]: pyro.subsample(
                                torch.zeros_like(self.latent_time_init),
                                event_dim=self.event_dim,
                            )
                        },
                    )
                else:
                    return autoguide.init_to_value(
                        site,
                        values={
                            site["name"]: pyro.subsample(
                                self.latent_time_init, event_dim=self.event_dim
                            )
                        },
                    )
            if site["name"] == "cell_gene_state":
                return autoguide.init_to_value(
                    site,
                    values={
                        site["name"]: pyro.subsample(
                            self.cell_gene_state_init, event_dim=self.event_dim
                        )
                    },
                )
            if site["name"] == "cell_time":
                assert getattr(self, f"{site['name']}_init").shape[-1] == 1
                print("guide init-------")
                print(self.cell_time_init.shape)
                print(
                    pyro.subsample(
                        self.cell_time_init, event_dim=self.event_dim
                    ).shape
                )
                print(self.event_dim)
                return autoguide.init_to_value(
                    site,
                    values={
                        site["name"]: pyro.subsample(
                            self.cell_time_init, event_dim=self.event_dim
                        )
                    },
                )
            if site["name"] == "t_scale":
                t_scale_init = self.alpha_init.new_ones(self.num_genes)
                return autoguide.init_to_value(
                    site, values={site["name"]: t_scale_init}
                )
            if site["name"] == "genecellpair_type":
                p_gene = self.alpha_init.new_ones(self.num_genes)
                return autoguide.init_to_value(
                    site, values={site["name"]: p_gene}
                )
            if site["name"] == "p_velocity":
                p_velocity = self.alpha_init.new_ones(self.num_genes)
                return autoguide.init_to_value(
                    site, values={site["name"]: p_velocity}
                )
            if hasattr(self, f"{site['name']}_init"):
                return autoguide.init_to_value(
                    site,
                    values={
                        site["name"]: getattr(self, f"{site['name']}_init")
                    },
                )
        return super().init(site)

    @PyroParam(constraint=positive, event_dim=0)
    def cell_code_scale(self):
        # return self.u_inf_init.new_ones(1)
        return self.u_inf.new_ones(1)

    def guide(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
    ):
        if self.plate_size == 2:
            cell_plate = self.plate(
                "cells", self.num_cells, subsample=ind_x, dim=-2
            )
            gene_plate = self.plate("genes", self.num_genes, dim=-1)
            if self.latent_factor == "linear":
                decoder_weights = self.map_estimate("cell_codebook")
                encoder_weights = decoder_weights.T

            with gene_plate:
                alpha = self.map_estimate("alpha")
                beta = self.map_estimate("beta")
                gamma = self.map_estimate("gamma")
                u_scale = self.map_estimate("u_scale")
                s_scale = self.map_estimate("s_scale")
                scale = u_scale / s_scale
                u_inf = self.u_inf
                s_inf = self.s_inf
                # u_inf = self.map_estimåte("u_inf")
                # s_inf = self.map_estimate("s_inf")
                # gamma = alpha / s_inf
                # beta = alpha / u_inf
                # pyro.sample("gamma", Delta(gamma))
                # pyro.sample("beta", Delta(beta))
                if self.t_scale_on:
                    t_scale = self.map_estimate("t_scale")
                switching = tau_inv(
                    u_inf, s_inf, self.zero, self.zero, alpha, beta, gamma
                )
                pyro.sample("switching", Delta(switching))

                if self.latent_factor == "linear":
                    u_pcs_mean = self.map_estimate("u_pcs_mean")
                    s_pcs_mean = self.map_estimate("s_pcs_mean")

            if not (self.likelihood in ["Normal", "LogNormal"]):
                # with cell_plate:
                #    u_read_depth = pyro.sample('u_read_depth', LogNormal(u_log_library, self.one.to(u_log_library.device)*0.01))
                #    s_read_depth = pyro.sample('s_read_depth', LogNormal(s_log_library, self.one.to(u_log_library.device)*0.01))
                with cell_plate, gene_plate, poutine.mask(
                    mask=pyro.subsample(
                        self.mask.to(u_obs.device), event_dim=self.event_dim
                    )
                ):
                    # u_s_loc = self.neuralnetwork(torch.cat([u_loc, s_loc], dim=-1))
                    # u_loc, s_loc = u_s_loc[..., :self.num_genes], u_s_loc[..., self.num_genes:]
                    # assert u_loc.shape == (u_obs.shape[0], self.num_genes)
                    # ut = pyro.sample("ut", Delta(torch.exp(u_loc).clamp(max=1e6)))
                    # st = pyro.sample("st", Delta(torch.exp(s_loc).clamp(max=1e6)))
                    ut = u_obs  # / u_read_depth
                    st = s_obs  # / s_read_depth
                    # ut = torch.log1p(u_obs / u_read_depth)
                    # st = torch.log1p(s_obs / s_read_depth)

            if self.shared_time:
                with cell_plate:
                    cell_time = self.map_estimate("cell_time")

            if self.latent_factor == "linear":
                cell_code_scale = pyro.param(
                    "cell_code_scale",
                    lambda: torch.tensor(0.1),
                    constraint=positive,
                ).to(u_obs.device)
                with cell_plate:
                    cell_code_loc = (
                        torch.cat(
                            (u_obs - u_pcs_mean, s_obs - s_pcs_mean), dim=-1
                        )
                        @ encoder_weights
                    )
                    cell_code_loc = cell_code_loc.unsqueeze(-1).transpose(
                        -1, -2
                    )
                    cell_code = pyro.sample(
                        "cell_code",
                        Normal(cell_code_loc, cell_code_scale).to_event(1),
                    )

            with cell_plate, gene_plate, poutine.mask(
                mask=pyro.subsample(self.mask.to(alpha.device), event_dim=0)
            ):
                if self.latent_factor == "linear":
                    regressor_output = torch.einsum(
                        "abc,cd->ad", cell_code, decoder_weights.squeeze()
                    )
                    regressor_u = softplus(
                        regressor_output[..., : self.num_genes].squeeze()
                        + u_pcs_mean
                    )
                    regressor_s = softplus(
                        regressor_output[..., self.num_genes :].squeeze()
                        + s_pcs_mean
                    )
                if self.latent_factor_operation == "sum":
                    u_obs = (u_obs - regressor_u).clamp(0.0)
                    s_obs = (s_obs - regressor_s).clamp(0.0)

                if not (self.likelihood in ["Normal", "LogNormal"]):
                    u_ = ut / scale
                    s_ = st
                else:
                    u_ = u_obs / scale
                    s_ = s_obs

                tau = tau_inv(u_, s_, self.zero, self.zero, alpha, beta, gamma)
                ut, st = mRNA(tau, self.zero, self.zero, alpha, beta, gamma)
                tau_ = tau_inv(u_, s_, u_inf, s_inf, self.zero, beta, gamma)
                ut_, st_ = mRNA(tau_, u_inf, s_inf, self.zero, beta, gamma)
                std_u = u_scale / scale
                state_on = ((ut - u_) / std_u) ** 2 + ((st - s_) / s_scale) ** 2
                state_off = ((ut_ - u_) / std_u) ** 2 + (
                    (st_ - s_) / s_scale
                ) ** 2
                cell_gene_state_logits = state_on - state_off
                # debug(cell_gene_state_logits)
                state = (
                    pyro.sample(
                        "cell_gene_state",
                        Bernoulli(logits=cell_gene_state_logits).to_event(
                            self.event_dim
                        ),
                    )
                    == self.zero
                )
                t = torch.where(state, tau, tau_ + switching)
                if self.shared_time:
                    if self.t_scale_on:
                        t = t - cell_time * t_scale
                    else:
                        t = t - cell_time
                pyro.sample("latent_time", Delta(t))

                if self.latent_factor_operation == "selection":
                    # p_velocity = self.map_estimate('p_velocity')
                    ut = torch.where(state, ut, ut_)
                    st = torch.where(state, st, st_)
                    velocity_model = ((ut - u_) / std_u) ** 2 + (
                        (st - s_) / s_scale
                    ) ** 2
                    pca_model = ((regressor_u / scale - u_) / std_u) ** 2 + (
                        (regressor_s - s_) / s_scale
                    ) ** 2
                    p_gene_type = pca_model - velocity_model  # .sum(axis=-2)
                    gene_type = pyro.sample(
                        "genecellpair_type", Bernoulli(logits=p_gene_type)
                    )
        else:
            self.guide2(u_obs, s_obs, u_log_library, s_log_library, ind_x)

    def guide2(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
    ):
        alpha = self.map_estimate("alpha")
        beta = self.map_estimate("beta")
        gamma = self.map_estimate("gamma")
        u_scale = self.map_estimate("u_scale")
        s_scale = self.map_estimate("s_scale")
        scale = u_scale / s_scale
        u_inf = pyro.param(
            "u_inf",
            lambda: self.u_inf_init.clone(),
            constraint=positive,
            event_dim=self.event_dim,
        ).to(u_obs.device)
        s_inf = pyro.param(
            "s_inf",
            lambda: self.s_inf_init.clone(),
            constraint=positive,
            event_dim=self.event_dim,
        ).to(u_obs.device)
        switching = tau_inv(
            u_inf, s_inf, self.zero, self.zero, alpha, beta, gamma
        )
        pyro.sample("switching", Delta(switching).to_event(1))

        cell_plate = self.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-1
        )
        with cell_plate:  # , poutine.mask(mask=pyro.subsample(self.mask.to(u_obs.device), event_dim=self.event_dim)):
            u_ = u_obs / scale
            s_ = s_obs
            tau = tau_inv(u_, s_, self.zero, self.zero, alpha, beta, gamma)
            ut, st = mRNA(tau, self.zero, self.zero, alpha, beta, gamma)
            tau_ = tau_inv(u_, s_, u_inf, s_inf, self.zero, beta, gamma)
            ut_, st_ = mRNA(tau_, u_inf, s_inf, self.zero, beta, gamma)
            std_u = u_scale / scale
            state_on = ((ut - u_) / std_u) ** 2 + ((st - s_) / s_scale) ** 2
            state_off = ((ut_ - u_) / std_u) ** 2 + ((st_ - s_) / s_scale) ** 2
            cell_gene_state_logits = state_on - state_off
            # debug(cell_gene_state_logits)
            state = (
                pyro.sample(
                    "cell_gene_state",
                    Bernoulli(logits=cell_gene_state_logits).to_event(
                        self.event_dim
                    ),
                )
                == self.zero
            )
            t = torch.where(state, tau, tau_ + switching)
            pyro.sample("latent_time", Delta(t).to_event(self.event_dim))


class AuxCellVelocityGuide(VelocityGuide):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.infer_cell = CellGuide(
            self.num_cells,
            self.num_genes,
            self.cell_time_init,
            self.zero,
            self.one,
            self.shared_time,
            self.likelihood,
            self.latent_factor,
            self.latent_factor_operation,
            self.t_scale_on,
        )
        if self.num_aux_cells > 0:
            self.infer_aux_cell = CellGuide(
                self.num_aux_cells,
                self.num_genes,
                self.cell_time_aux_init,
                self.zero,
                self.one,
                self.shared_time,
                self.likelihood,
                self.latent_factor,
                self.latent_factor_operation,
                self.t_scale_on,
            )

    def guide(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_log_library_loc: Optional[torch.Tensor] = None,
        s_log_library_loc: Optional[torch.Tensor] = None,
        u_log_library_scale: Optional[torch.Tensor] = None,
        s_log_library_scale: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
    ):
        cell_plate = self.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-2
        )
        gene_plate = self.plate("genes", self.num_genes, dim=-1)

        with gene_plate:
            alpha = self.map_estimate("alpha")
            beta = self.map_estimate("beta")
            gamma = self.map_estimate("gamma")
            u_scale = self.map_estimate("u_scale")
            s_scale = self.map_estimate("s_scale")
            u0 = self.map_estimate("u_offset")
            s0 = self.map_estimate("s_offset")

            # if self.t_scale_on and self.shared_time:
            #    t_scale = self.map_estimate("t_scale")
            # else:
            #    t_scale = None
            #    gene_offset = None
            # alpha, beta, gamma = AutoDiagonalNormal()

            t_scale = None
            # u0 = s0 = 0
            # t_scale = self.map_estimate("t_scale")
            if self.add_offset:
                gene_offset = self.map_estimate("gene_offset")
            else:
                gene_offset = 0

            if self.only_cell_times:
                dt_switching = self.map_estimate("dt_switching")
                u_inf, s_inf = mRNA(
                    dt_switching, u0, s0, self.alpha, self.beta, self.gamma
                )

                ##s_inf/u_inf from alpha/beta/gamma computation
                ##leads all genes to be uniphases, perhaps due to initialization?
                ##with slightly better shared time prediction in dentate gyrus
                # s_inf = alpha / gamma
                # u_inf = alpha / beta
                ##u0 = torch.where(u0 > u_inf, u_inf, u0)
                ##s0 = torch.where(s0 > s_inf, s_inf, s0)
                # dt_switching = tau_inv(u_inf, s_inf, u0, s0, alpha, beta, gamma)
                # dt_switching = pyro.sample("dt_switching", Delta(dt_switching))
            else:
                u_inf = pyro.param(
                    "u_inf",
                    lambda: self.u_inf_init.clone(),
                    constraint=positive,
                    event_dim=self.event_dim,
                ).to(u_obs.device)
                s_inf = pyro.param(
                    "s_inf",
                    lambda: self.s_inf_init.clone(),
                    constraint=positive,
                    event_dim=self.event_dim,
                ).to(u_obs.device)
                switching = tau_inv(
                    u_inf, s_inf, self.zero, self.zero, alpha, beta, gamma
                )
                pyro.sample("switching", Delta(switching))

            if self.latent_factor == "linear":
                u_pcs_mean = self.map_estimate("u_pcs_mean")
                s_pcs_mean = self.map_estimate("s_pcs_mean")
                if self.latent_factor_operation == "selection":
                    p_velocity = self.map_estimate("p_velocity")
                else:
                    p_velocity = None
            else:
                u_pcs_mean = None
                s_pcs_mean = None
                p_velocity = None

        # with cell_plate, gene_plate:
        #    cell_gene_mask = pyro.subsample(self.mask.to(alpha.device), event_dim=0)

        if self.latent_factor == "linear":
            decoder_weights = self.map_estimate("cell_codebook")
        else:
            decoder_weights = None

        # same as before, just refactored a bit
        with cell_plate:  # , poutine.mask(mask=cell_gene_mask):
            u_read_depth = None
            s_read_depth = None
            ut, st, u_, s_, regressor_u, regressor_s = self.infer_cell(
                gene_plate,
                u_obs,
                s_obs,
                u_log_library,
                s_log_library,
                alpha,
                beta,
                gamma,
                # u_inf, s_inf, switching, u_scale, s_scale,
                u_inf,
                s_inf,
                dt_switching,
                u_scale,
                s_scale,
                u_pcs_mean,
                s_pcs_mean,
                t_scale,
                decoder_weights,
                gene_offset,
                p_velocity,
                self.only_cell_times,
                u_read_depth=u_read_depth,
                s_read_depth=s_read_depth,
                u0=u0,
                s0=s0,
            )  # TODO arguments and return type
        if self.num_aux_cells > 0:
            # new: infer latent variables for auxiliary cells
            with pyro.contrib.autoname.scope(prefix="aux"):
                # TODO set self.num_aux_cells
                aux_cell_plate = self.plate(
                    "aux_cell_plate", self.num_aux_cells, dim=cell_plate.dim
                )
                with aux_cell_plate:
                    aux_u_obs = pyro.param(
                        "aux_u_obs",
                        lambda: self.aux_u_obs_init,
                        constraint=positive,
                        event_dim=0,
                    )  # TODO initialize
                    aux_s_obs = pyro.param(
                        "aux_s_obs",
                        lambda: self.aux_s_obs_init,
                        constraint=positive,
                        event_dim=0,
                    )  # TODO initialize
                    aux_u_log_library = torch.log(
                        aux_u_obs.sum(axis=-1) + 1e-6
                    )  # TODO define
                    aux_s_log_library = torch.log(
                        aux_s_obs.sum(axis=-1) + 1e-6
                    )  # TODO define
                    (
                        aux_ut,
                        aux_st,
                        aux_u_,
                        aux_s_,
                        aux_regressor_u,
                        aux_regressor_s,
                    ) = self.infer_aux_cell(
                        gene_plate,
                        aux_u_obs,
                        aux_s_obs,
                        aux_u_log_library,
                        aux_s_log_library,
                        alpha,
                        beta,
                        gamma,
                        # u_inf, s_inf, switching, u_scale, s_scale,
                        u_inf,
                        s_inf,
                        dt_switching,
                        u_scale,
                        s_scale,
                        u_pcs_mean,
                        s_pcs_mean,
                        t_scale,
                        decoder_weights,
                        gene_offset,
                        p_velocity,
                        self.only_cell_times,
                        u_read_depth=u_read_depth,
                        s_read_depth=s_read_depth,
                        u0=u0,
                        s0=s0,
                    )  # TODO arguments and return type
            # if self.latent_factor_operation == 'selection':
            #    ut = torch.cat([ut, aux_ut], dim=cell_plate.dim)
            #    st = torch.cat([st, aux_st], dim=cell_plate.dim)
            #    u_ = torch.cat([u_, aux_u_], dim=cell_plate.dim)
            #    s_ = torch.cat([s_, aux_s_], dim=cell_plate.dim)
            #    regressor_u = torch.cat([regressor_u, aux_regressor_u], dim=cell_plate.dim)
            #    regressor_s = torch.cat([regressor_s, aux_regressor_s], dim=cell_plate.dim)

        # gene selection version
        # cell-gene selection should be moved into infer_cell
        # if self.latent_factor_operation == 'selection':
        #    with gene_plate, self.plate("cells", self.num_cells+self.num_aux_cells, subsample=ind_x, dim=-2):
        #        scale = u_scale / s_scale
        #        std_u = u_scale / scale
        #        velocity_error = ((ut - u_)/std_u) ** 2 + ((st - s_)/s_scale) ** 2
        #        pca_error = ((regressor_u/scale - u_)/std_u) ** 2 + ((regressor_s - s_)/s_scale) ** 2
        #        probs = torch.stack([
        #            torch.exp(-velocity_error) - p_velocity,
        #            torch.exp(-pca_error) - (1 - p_velocity)
        #        ], dim=-1)
        #        probs = softmax(probs, dim=-1)  # probs.sum(-1, keepdim=True)
        #        p_gene_type = probs[..., 0]
        #        gene_type = pyro.sample("cellgene_type", Bernoulli(p_gene_type)) ## tend to be constant


class MultiKineticsGuide(VelocityGuide):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.latent_kinetics_comp = 10
        ##option 1: encoder for latent space of alpha/beta/gamma
        # self.us_encoder = Encoder(
        #    self.num_genes * 2,
        #    n_output=self.latent_kinetics_comp,
        #    activation_fn=nn.ELU,
        #    n_layers=2, var_eps=1e-6)
        ##option 1: decoder of alpha/beta/gamma
        # self.infer_kinetics = Decoder(
        #    self.latent_kinetics_comp,
        #    self.num_genes*3, # * 6,
        #    n_layers=2,
        #    activation_fn=nn.ELU,
        #    n_hidden=128,
        # )
        self.time_encoder = TimeEncoder(
            self.num_genes,
            n_output=1,
            activation_fn=nn.ELU,
            n_layers=2,
            var_eps=1e-6,
        )
        self.k = 2

    def guide(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_log_library_loc: Optional[torch.Tensor] = None,
        s_log_library_loc: Optional[torch.Tensor] = None,
        u_log_library_scale: Optional[torch.Tensor] = None,
        s_log_library_scale: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
        cell_state: Optional[torch.Tensor] = None,
        time_info: Optional[torch.Tensor] = None,
    ):
        # pyro.module("gamma_celltime_encoder", self.gamma_celltime_encoder)
        # pyro.module("infer_kinetics", self.infer_kinetics)
        # pyro.module("us_encoder", self.us_encoder)
        pyro.module("time_encoder", self.time_encoder)
        cell_plate = self.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-2
        )
        gene_plate = self.plate("genes", self.num_genes, dim=-1)
        kinetics_plate = self.plate("kinetics", self.k, dim=-3)
        with kinetics_plate, gene_plate:
            alpha_k = self.map_estimate("alpha")
            beta_k = self.map_estimate("beta")
            gamma_k = self.map_estimate("gamma")
            dt_switching_k = self.map_estimate("dt_switching")
            # t0_k = self.map_estimate('t0')
            # u_offset_k = self.map_estimate('u_offset')
            # s_offset_k = self.map_estimate('s_offset')
            # u_scale = self.map_estimate('u_scale')
            # s_scale = self.map_estimate('s_scale')

        with gene_plate:
            u0 = s0 = self.zero
            u_scale = self.map_estimate("u_scale")
            s_scale = self.map_estimate("s_scale")
            t0 = self.map_estimate("t0")
            u0 = u_offset = self.map_estimate("u_offset")
            s0 = s_offset = self.map_estimate("s_offset")

        # alpha = alpha_k[cluster_ind].squeeze()
        # beta = beta_k[cluster_ind].squeeze()
        # gamma = gamma_k[cluster_ind].squeeze()
        # t0 = t0_k[cluster_ind].squeeze()
        # u0 = u_offset = u_offset_k[cluster_ind].squeeze()
        # s0 = s_offset = s_offset_k[cluster_ind].squeeze()
        # u_offset = u_offset_k[cluster_ind].squeeze()
        # s_offset = s_offset_k[cluster_ind].squeeze()
        # u_scale = u_scale[cluster_ind].squeeze()
        # s_scale = s_scale[cluster_ind].squeeze()

        # dt_switching = dt_switching_k[cluster_ind].squeeze()
        # u_inf, s_inf = mRNA(dt_switching, u0, s0, alpha, beta, gamma)
        # switching = t0 + dt_switching

        # with cell_plate:
        #    latent_kinetics_loc, latent_kinetics_scale, z = self.us_encoder(torch.hstack([u_obs / (self.u_scale / self.s_scale), s_obs]))
        # kinetics_params, _ = self.infer_kinetics(latent_kinetics_loc.squeeze())
        # alpha = softplus(kinetics_params[:, :self.num_genes])
        # beta = softplus(kinetics_params[:, self.num_genes:(self.num_genes*2)])
        # gamma = softplus(kinetics_params[:, (self.num_genes*2):(self.num_genes*3)])
        ##t0 = kinetics_params[:, (self.num_genes*3):(self.num_genes*4)]
        ##u0 = softplus(kinetics_params[:, (self.num_genes*4):(self.num_genes*5)])
        ##s0 = softplus(kinetics_params[:, (self.num_genes*5):(self.num_genes*6)])
        # with cell_plate, gene_plate:
        #    pyro.sample("alpha", Delta(alpha, event_dim=0))
        #    pyro.sample("beta", Delta(beta, event_dim=0))
        #    pyro.sample("gamma", Delta(gamma, event_dim=0))
        # pyro.sample("t0", Delta(t0, event_dim=0))
        # pyro.sample("u0", Delta(u0, event_dim=0))
        # pyro.sample("s0", Delta(s0, event_dim=0))

        scale = u_scale / s_scale
        s_ = s_obs
        u_ = u_obs / scale

        tau = tau_inv(u_, s_, u0, s0, alpha, beta, gamma)
        ut, st = mRNA(tau, u0, s0, alpha, beta, gamma)
        tau_ = tau_inv(u_, s_, u_inf, s_inf, self.zero, beta, gamma)
        ut_, st_ = mRNA(tau_, u_inf, s_inf, self.zero, beta, gamma)
        std_u = u_scale / scale
        state_on = ((ut - u_) / std_u) ** 2 + ((st - s_) / s_scale) ** 2
        state_off = ((ut_ - u_) / std_u) ** 2 + ((st_ - s_) / s_scale) ** 2
        state_zero = ((ut - u0) / std_u) ** 2 + ((st - s0) / s_scale) ** 2
        state_inf = ((ut_ - u_inf) / std_u) ** 2 + (
            (st_ - s_inf) / s_scale
        ) ** 2
        cell_gene_state_logits = torch.stack(
            [state_on, state_zero, state_off, state_inf], dim=-1
        ).argmin(-1)
        state = (cell_gene_state_logits > 1) == self.zero
        t = torch.where(state, tau + t0, tau_ + switching)
        ##t = tau + t0
        cell_time_loc = self.time_encoder(t)
        with cell_plate:
            pyro.sample("cell_time", Delta(cell_time_loc, event_dim=0))


class TimeEncoder(nn.Module):
    """adapt https://docs.scvi-tools.org/en/0.9.1/_modules/scvi/nn/_base_components.html#Encoder
    for shap DeepExplainer usage
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        n_cat_list: Iterable[int] = None,
        n_layers: int = 1,
        n_hidden: int = 128,
        dropout_rate: float = 0.1,
        distribution: str = "normal",
        var_eps: float = 1e-6,
        **kwargs,
    ):
        super().__init__()
        # self.distribution = distribution
        self.var_eps = var_eps
        self.encoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            **kwargs,
        )
        self.mean_encoder = nn.Linear(n_hidden, n_output)
        self.act = nn.Softplus(beta=1)
        # self.var_encoder = nn.Linear(n_hidden, n_output)
        # if distribution == "ln":
        #    self.z_transformation = nn.Softmax(dim=-1)
        # else:
        #    self.z_transformation = identity

    # def forward(self, x: torch.Tensor, *cat_list: int, u_log_library: torch.Tensor, s_log_library: torch.Tensor):
    def forward(self, x: torch.Tensor, *cat_list: int):
        r"""
        The forward computation for a single sample.
         #. Encodes the data into latent space using the encoder network
         #. Generates a mean \\( q_m \\) and variance \\( q_v \\)
         #. Samples a new value from an i.i.d. multivariate normal \\( \\sim Ne(q_m, \\mathbf{I}q_v) \\)
        Parameters
        ----------
        x
            tensor with shape (n_input,)
        cat_list
            list of category membership(s) for this sample
        Returns
        -------
        3-tuple of :py:class:`torch.Tensor`
            tensors of shape ``(n_latent,)`` for mean and var, and sample
        """
        # Parameters for latent distribution
        # x = torch.hstack([x, u_log_library, s_log_library]) # insert library sizes

        q = self.encoder(x, *cat_list)
        q_m = self.act(
            self.mean_encoder(q)
        )  # really big cell_time, this also generates reverse order with poisson
        # q_m = self.mean_encoder(q) # really big cell_time, this also generates reverse order with poisson
        # q1 = self.act(q_m[:, 0].reshape((-1, 1)))
        # q2 = q_m[:, 1].reshape((-1, 1))
        # q3 = self.act(q_m[:, 2].reshape((-1, 1)))
        # q4 = q_m[:, 3].reshape((-1, 1))
        # q5 = self.act(q_m[:, 4].reshape((-1, 1)))
        # q_m = self.mean_encoder(q)           # huge cell_time, this generates reverse order
        # q_m = torch.tanh(self.mean_encoder(q)) # contrain not work
        # q_v = torch.exp(self.var_encoder(q)) + self.var_eps # This pretty poor performance
        # q_v = nn.ReLU()(self.var_encoder(q)) + self.var_eps # nan values
        # q_v = self.act(self.var_encoder(q)) + self.var_eps  # slightly better..
        # latent = self.z_transformation(reparameterize_gaussian(q_m, q_v))
        return q_m  # , q_v
        # return q1, q2, q3, q4, q5


class CellGuide(PyroModule):
    def __init__(
        self,
        num_cells,
        num_genes,
        init_cell_time,
        zero,
        one,
        shared_time,
        likelihood,
        latent_factor,
        latent_factor_operation,
        t_scale_on,
    ):
        super().__init__()
        self.num_cells = num_cells  # TODO this shouldn't be required
        self.num_genes = num_genes
        self.shared_time = shared_time
        self.latent_factor = latent_factor
        self.latent_factor_operation = latent_factor_operation
        self.t_scale_on = t_scale_on
        if self.shared_time:
            if not (init_cell_time is None):
                init_cell_time = (
                    init_cell_time.detach().clone().requires_grad_()
                )
                self.cell_time_loc = PyroParam(init_cell_time, event_dim=0)
        self.cell_code_scale = PyroParam(
            torch.tensor(0.1), constraint=positive, event_dim=0
        )
        self.zero = zero
        self.one = one
        self.u_loc = PyroParam(
            torch.ones(self.num_cells, self.num_genes),
            constraint=positive,
            event_dim=0,
        )
        self.s_loc = PyroParam(
            torch.ones(self.num_cells, self.num_genes),
            constraint=positive,
            event_dim=0,
        )
        self.likelihood = likelihood
        self.encoder = TimeEncoder(
            self.num_genes,  # +2 for U and S library sizes
            n_output=1,
            dropout_rate=0.5,
            activation_fn=nn.ELU,
            n_layers=3,
            var_eps=1e-6,
        )

    def forward(
        self,
        gene_plate,
        u_obs,
        s_obs,
        u_log_library,
        s_log_library,
        alpha,
        beta,
        gamma,
        # u_inf, s_inf, switching, u_scale, s_scale,
        u_inf,
        s_inf,
        dt_switching,
        u_scale,
        s_scale,
        u_pcs_mean,
        s_pcs_mean,
        t_scale,
        decoder_weights,
        gene_offset,
        p_velocity=None,
        only_cell_times=False,
        u_read_depth=None,
        s_read_depth=None,
        u0=None,
        s0=None,
    ):  # TODO arguments and return type
        pyro.module("time_encoder", self.encoder)
        scale = u_scale / s_scale
        if self.latent_factor == "linear":
            encoder_weights = decoder_weights.T
            cell_code_loc = (
                torch.cat((u_obs - u_pcs_mean, s_obs - s_pcs_mean), dim=-1)
                @ encoder_weights
            )
            cell_code_loc = cell_code_loc.unsqueeze(-1).transpose(-1, -2)
            cell_code = pyro.sample(
                "cell_code",
                Normal(cell_code_loc, self.cell_code_scale).to_event(1),
            )
        with gene_plate:
            if self.latent_factor == "linear":
                regressor_output = torch.einsum(
                    "abc,cd->ad", cell_code, decoder_weights.squeeze()
                )
                regressor_u = softplus(
                    regressor_output[..., : self.num_genes].squeeze()
                    + u_pcs_mean
                )
                regressor_s = softplus(
                    regressor_output[..., self.num_genes :].squeeze()
                    + s_pcs_mean
                )
            if self.latent_factor_operation == "sum":
                u_obs = u_obs - regressor_u
                s_obs = s_obs - regressor_s
            if self.likelihood != "Normal":
                # u_ = torch.log1p(u_obs) / scale
                # s_ = torch.log1p(s_obs)
                u_ = u_obs / scale
                s_ = s_obs
            else:
                u_ = u_obs / scale
                s_ = s_obs

            tau = tau_inv(u_, s_, u0, s0, alpha, beta, gamma)
            # tau = torch.where(tau > dt_switching, dt_switching, tau)
            ut, st = mRNA(tau, u0, s0, alpha, beta, gamma)
            tau_ = tau_inv(u_, s_, u_inf, s_inf, self.zero, beta, gamma)
            ##tau_ = torch.where(tau_ >= tau_[s_obs > 0].max(dim=0)[0], tau_[s_obs > 0].max(dim=0)[0], tau_)
            ut_, st_ = mRNA(tau_, u_inf, s_inf, self.zero, beta, gamma)
            std_u = u_scale / scale
            state_on = ((ut - u_) / std_u) ** 2 + ((st - s_) / s_scale) ** 2
            state_off = ((ut_ - u_) / std_u) ** 2 + ((st_ - s_) / s_scale) ** 2
            state_zero = ((ut - u0) / std_u) ** 2 + ((st - s0) / s_scale) ** 2
            state_inf = ((ut_ - u_inf) / std_u) ** 2 + (
                (st_ - s_inf) / s_scale
            ) ** 2
            cell_gene_state_logits = torch.stack(
                [state_on, state_zero, state_off, state_inf], dim=-1
            ).argmin(-1)
            state = (cell_gene_state_logits > 1) == self.zero
            switching = dt_switching + gene_offset
            t = torch.where(state, tau + gene_offset, tau_ + switching)
            ut = torch.where(state, ut, ut_)
            st = torch.where(state, st, st_)
            # cell_gene_state_logits = state_on-state_off
            # state = pyro.sample("cell_gene_state", Bernoulli(logits=cell_gene_state_logits)) == self.zero
            # gene_selection_logits = (ut - u_)**2 + (st - s_)**2
            # gene_selection = pyro.sample("gene_selection", Bernoulli(logits=gene_selection_logits)) == self.zero
            ##gene_selection_logits = torch.argsort(gene_selection_logits.sum(-2))
            ##gene_selection = alpha.new_zeros(alpha.shape)
            ##gene_selection[gene_selection_logits[:300]] = u0 + 1
            # t = t * gene_selection
        if self.shared_time:
            if not only_cell_times:
                cell_time_loc = torch.mean(
                    ((t - gene_offset) / t_scale)
                    if self.t_scale_on
                    else t,  # (t - gene_offset),
                    dim=gene_plate.dim,
                    keepdim=True,
                )
                cell_time_dist = Delta(cell_time_loc, event_dim=0)
                cell_time = pyro.sample("cell_time", cell_time_dist)
            else:
                # cell_time_loc, cell_time_scale = self.encoder(t)
                cell_time_loc = self.encoder(t)
                # cell_time_loc, u_read_depth_loc, u_read_depth_scale, s_read_depth_loc, s_read_depth_scale = self.encoder(t, u_log_library=u_log_library, s_log_library=s_log_library)
                # u_read_depth = pyro.sample('u_read_depth', LogNormal(u_read_depth_loc, u_read_depth_scale))
                # s_read_depth = pyro.sample('s_read_depth', LogNormal(s_read_depth_loc, s_read_depth_scale))
                cell_time = pyro.sample(
                    "cell_time", Delta(cell_time_loc, event_dim=0)
                )

                # cell_time_dist = TransformedDistribution(
                #            Normal(cell_time_loc, torch.sqrt(cell_time_scale)),
                #            #ExpTransform())  # equivalent to LogNormal(cell_time_loc, cell_time_scale), spelled out to clarify
                #            SoftplusTransform())
                # cell_time = pyro.sample("cell_time", cell_time_dist)

        # if self.likelihood != 'Normal':
        #    pyro.sample("ut", Normal(ut * scale, u_scale))
        #    pyro.sample("st", Normal(st, s_scale))

        with gene_plate:
            if not only_cell_times:
                if self.shared_time:
                    if self.t_scale_on:
                        t = t - cell_time * t_scale - gene_offset
                    else:
                        t = t - cell_time  # - gene_offset # FIX!
                pyro.sample("latent_time", Delta(t))

            # model:
            # cell_time[c] ~ ...
            # switching[g] ~ ...
            # t_noise[c, g] ~ ...
            # t[c, g] = cell_time[c] - switching[g] + t_noise[c, g]
            # u_obs | t[c, g] ~ ...
            #
            # guide (currently):
            # cell_time[c] ~ ...
            # switching[g] ~ ...
            # t[c, g] = f(u_obs)  # constant wrt cell_time
            # t_noise[c, g] = t[c, g] - cell_time[c] + switching[g]
            #
            # thus when replaying the guide through the model:
            # t[c, g] = cell_time[c] - switching[g] + (f(u_obs) - cell_time[c] + switching)
            #         = f(u_obs)  # constant wrt cell_time

            # updating the guide
            # t[c, g] = f(u_obs)
            # cell_time[c] = ((t[c, g] + switching[g]) / t_scale[g]).mean(g)
            # t_noise[c, g] = t[c, g] - cell_time[c] * t_scale[g]

            # cell-gene selection
            if self.latent_factor_operation == "selection":
                scale = u_scale / s_scale
                std_u = u_scale / scale
                velocity_error = ((ut - u_) / std_u) ** 2 + (
                    (st - s_) / s_scale
                ) ** 2
                pca_error = ((regressor_u / scale - u_) / std_u) ** 2 + (
                    (regressor_s - s_) / s_scale
                ) ** 2
                probs = torch.stack(
                    [
                        torch.exp(-velocity_error) - p_velocity,
                        torch.exp(-pca_error) - (1 - p_velocity),
                    ],
                    dim=-1,
                )
                probs = softmax(probs, dim=-1)  # probs.sum(-1, keepdim=True)
                p_gene_type = probs[..., 0]
                cellgene_type = pyro.sample(
                    "cellgene_type", Bernoulli(p_gene_type)
                )  ## tend to be constant
            if (
                self.latent_factor_operation == "selection"
                and self.latent_factor == "linear"
            ):
                return ut, st, u_, s_, regressor_u, regressor_s
            return ut, st, u_, s_, None, None


class VelocityAutoGuideList(AutoGuideList):
    def forward(self, *args, **kwargs):
        """
        A composite guide with the same ``*args, **kwargs`` as the base ``model``.

        .. note:: This method is used internally by :class:`~torch.nn.Module`.
            Users should instead use :meth:`~torch.nn.Module.__call__`.

        :return: A dict mapping sample site name to sampled value.
        :rtype: dict
        """
        if self.model.guide_type in ["velocity_auto", "velocity_auto_depth"]:
            pyro.module("time_encoder", self.model.time_encoder)
        if self.model.correct_library_size:  # use autoguide for read depth
            if self.model.guide_type != "velocity_auto_depth":
                pyro.module("u_lib_encoder", self.model.u_lib_encoder)
                pyro.module("s_lib_encoder", self.model.s_lib_encoder)
        result = super().forward(*args, **kwargs)
        assert "alpha" in result
        assert "beta" in result
        assert "gamma" in result

        alpha = result["alpha"]
        beta = result["beta"]
        gamma = result["gamma"]

        if self.model.cell_specific_kinetics is not None:
            cell_state = args[-1]
            rho, _ = self.model.multikinetics_encoder(cell_state)
            alpha = rho * alpha

        if self.model.add_offset:
            u0 = result["u_offset"]
            s0 = result["s_offset"]
        else:
            u0 = s0 = alpha.new_zeros(alpha.shape)

        t0 = result["t0"]

        if "u_scale" in result:
            u_scale = result["u_scale"]
            s_scale = result["s_scale"]
        else:
            u_scale = s_scale = alpha.new_ones(alpha.shape)

        if self.model.cell_specific_kinetics is None:
            dt_switching = result["dt_switching"]
            u_inf, s_inf = mRNA(dt_switching, u0, s0, alpha, beta, gamma)
        else:
            dt_switching = None
            u_inf, s_inf = None, None

        with self.plates["cells"]:
            u_obs = args[0]
            s_obs = args[1]
            u_log_library = args[2]
            s_log_library = args[3]
            if self.model.correct_library_size:
                # if (self.model.guide_type != 'velocity_auto_depth'):
                #    u_lib_loc, u_lib_var = self.model.u_lib_encoder(torch.hstack([torch.log1p(u_obs), u_log_library]))
                #    s_lib_loc, s_lib_var = self.model.s_lib_encoder(torch.hstack([torch.log1p(s_obs), s_log_library]))
                #    result['u_read_depth'] = pyro.sample("u_read_depth", LogNormal(u_lib_loc, u_lib_var).to_event(0))
                #    result['s_read_depth'] = pyro.sample("s_read_depth", LogNormal(s_lib_loc, s_lib_var).to_event(0))
                if self.model.guide_type == "velocity_auto":
                    u_read_depth = None
                    s_read_depth = None
                if self.model.guide_type == "velocity_auto_depth":
                    u_read_depth = result["u_read_depth"]
                    s_read_depth = result["s_read_depth"]
            else:
                u_read_depth = s_read_depth = None
            t = self.model.get_time(
                u_scale,
                s_scale,
                alpha,
                beta,
                gamma,
                u_obs,
                s_obs,
                u0,
                s0,
                t0,
                dt_switching,
                u_inf,
                s_inf,
                u_read_depth=u_read_depth,
                s_read_depth=s_read_depth,
            )
            result["cell_time"] = t
        return result


class TrajectoryGuide(VelocityGuide):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def guide(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
    ):
        cell_plate = self.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-2
        )
        aux_cell_plate = self.plate("aux_cells", self.num_aux_cells, dim=-2)
        gene_plate = self.plate("genes", self.num_genes, dim=-1)

        with gene_plate, poutine.mask(mask=True):
            alpha = self.map_estimate("alpha")
            beta = self.map_estimate("beta")
            gamma = self.map_estimate("gamma")
            switching = self.map_estimate("switching")
            u_scale = self.map_estimate("u_scale")
            s_scale = self.map_estimate("s_scale")

        dt_aux = []
        u_aux = []
        s_aux = []
        zero = self.zero.to(u_obs.device)
        # u0 = s0 = zero
        u_prev = s_prev = zero

        for step in pyro.markov(range(aux_cell_plate.size)):
            dt_aux.append(self.map_estimate(f"dt_aux_{step}"))
            with gene_plate:
                u_noise = self.map_estimate(f"u_aux_noise_{step}")
                s_noise = self.map_estimate(f"s_aux_noise_{step}")
                state_aux = (
                    pyro.sample(
                        f"cell_gene_state_{step}",
                        Bernoulli(logits=dt_aux[step] - switching),
                    )
                    == zero
                )
                alpha = torch.where(state_aux, alpha, zero)
                u_loc, s_loc = mRNA(
                    dt_aux[step], u_prev, s_prev, alpha, beta, gamma
                )
                u_prev = u_loc + u_noise * torch.sqrt(2 * dt_aux[step])
                s_prev = s_loc + s_noise * torch.sqrt(2 * dt_aux[step])
                u_aux.append(u_prev)
                s_aux.append(s_prev)

        dt_aux = torch.hstack(dt_aux).to(u_obs.device)
        u_aux = torch.vstack(u_aux).to(u_obs.device)
        s_aux = torch.vstack(s_aux).to(u_obs.device)
        with cell_plate:
            # order_stat = torch.zeros(u_obs.shape[0], 1, u_aux.shape[0]).to(u_obs.device)
            # for i in range(order_stat.shape[0]):
            #    for j in range(order_stat.shape[1]):
            #        order_stat[i,0,j] = ((u_obs[i] - u_aux[j])**2+(s_obs[i] - s_aux[j])**2).sum()
            order_stat = (
                (u_obs.reshape(u_obs.shape[0], 1, u_obs.shape[1]) - u_aux) ** 2
            ).sum() + (
                (s_obs.reshape(s_obs.shape[0], 1, s_obs.shape[1]) - s_aux) ** 2
            ).sum(axis=-1)
            order_stat = order_stat.reshape(u_obs.shape[0], 1, u_aux.shape[0])
            order = pyro.sample(
                "order",
                Categorical(logits=-order_stat),
                infer={"enumerate": "parallel"},
            )
            dt = dt_aux[..., order]
            with gene_plate:
                u_noise = self.map_estimate("u_noise")
                s_noise = self.map_estimate("s_noise")
                u_prev = u_aux[..., order, :].squeeze()
                s_prev = s_aux[..., order, :].squeeze()
                u_loc_on, s_loc_on = mRNA(
                    dt, u_prev, s_prev, alpha, beta, gamma
                )
                u_loc_off, s_loc_off = mRNA(
                    dt, u_prev, s_prev, zero, beta, gamma
                )
                u_on = u_loc_on + u_noise * torch.sqrt(2 * dt)
                s_on = s_loc_on + s_noise * torch.sqrt(2 * dt)
                u_off = u_loc_off + u_noise * torch.sqrt(2 * dt)
                s_off = s_loc_off + s_noise * torch.sqrt(2 * dt)
                scale = u_scale / s_scale
                std_u = u_scale / scale
                mse_on = ((u_on - u_obs / scale) / std_u) ** 2 + (
                    (s_on - s_obs) / s_scale
                ) ** 2
                mse_off = ((u_off - u_obs / scale) / std_u) ** 2 + (
                    (s_off - s_obs) / s_scale
                ) ** 2
                state = pyro.sample(
                    "cell_gene_state", Bernoulli(logits=mse_on - mse_off)
                )


class DecoderTimeGuide(VelocityGuide):
    def __init__(self, model: VelocityModel):
        super().__init__(model)
        self.encoder = Encoder(self.num_genes * 2, n_output=10, n_layers=1)

    def guide(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
    ):
        pyro.module("time_encoder", self.encoder)
        gene_plate = self.plate("genes", self.num_genes, dim=-1)
        cell_plate = self.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-2
        )
        u_ = torch.log1p(u_obs)
        s_ = torch.log1p(s_obs)
        x = torch.hstack([u_, s_])

        with gene_plate:
            switching = self.map_estimate("switching")
            alpha = self.map_estimate("alpha")
            beta = self.map_estimate("beta")
            gamma = self.map_estimate("gamma")

        with cell_plate:
            z_loc, z_scale, _ = self.encoder(x)
            z_loc = z_loc.unsqueeze(-2)
            z_scale = z_scale.unsqueeze(-2)
            pyro.sample(
                "latent_time_latent_space", Normal(z_loc, z_scale).to_event(1)
            )  # (cell, 1, components)


class KineticsParamEncoder(nn.Module):
    def __init__(self, hidden_shape, cell_size):
        super().__init__()
        self.network = nn.Sequential(
            nn.Linear(2, hidden_shape),  # 2 for (u, s)
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Linear(hidden_shape, 4),
            nn.ReLU(),
        )
        self.linear = nn.Sequential(nn.Linear(cell_size, 1), nn.Softplus())
        # self.u_inf = nn.Sequential(
        #    nn.Linear(cell_size, 1),
        #    nn.Softplus()
        # )
        # self.s_inf = nn.Sequential(
        #    nn.Linear(cell_size, 1),
        #    nn.Softplus()
        # )

    def forward(self, u_obs, s_obs):
        kinetics_param = self.network(torch.stack([u_obs, s_obs], dim=-1))
        alpha = kinetics_param[..., 0].T
        beta = kinetics_param[..., 1].T
        gamma = kinetics_param[..., 2].T
        t = kinetics_param[..., 3].T

        alpha = self.linear(alpha).squeeze(-1)
        beta = self.linear(beta).squeeze(-1)
        gamma = self.linear(gamma).squeeze(-1)
        switching = self.linear(t).squeeze(-1)

        # u_inf = self.u_inf(u_obs.T).squeeze(-1)
        # s_inf = self.s_inf(s_obs.T).squeeze(-1)
        # return alpha, beta, gamma, u_inf, s_inf
        return alpha, beta, gamma, switching, softplus(t.T)


class GeneTimeEncoder(nn.Module):
    def __init__(self, hidden_shape):
        super().__init__()
        self.network = nn.Sequential(
            nn.Linear(7, hidden_shape),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Linear(hidden_shape, hidden_shape),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Linear(hidden_shape, 1),
        )

    def forward(
        self,
        u_obs,
        s_obs,
        alpha,
        alpha_off,
        beta,
        gamma,
        u0,
        s0,
        u_inf=None,
        s_inf=None,
    ):
        switching = self.network(
            torch.stack(
                [
                    u_inf,
                    s_inf,
                    u0.expand_as(u_inf),
                    s0.expand_as(s_inf),
                    alpha,
                    beta,
                    gamma,
                ],
                dim=-1,
            )
        )
        switching = switching[..., 0]
        tau = self.network(
            torch.stack(
                [
                    u_obs,
                    s_obs,
                    u0.expand_as(u_obs),
                    s0.expand_as(s_obs),
                    alpha.expand_as(u_obs),
                    beta.expand_as(s_obs),
                    gamma.expand_as(s_obs),
                ],
                dim=-1,
            )
        )
        tau = tau[..., 0]

        tau_ = self.network(
            torch.stack(
                [
                    u_obs,
                    s_obs,
                    u_inf.expand_as(u_obs),
                    s_inf.expand_as(s_obs),
                    torch.tensor(alpha_off).expand_as(u_obs),
                    beta.expand_as(u_obs),
                    gamma.expand_as(u_obs),
                ],
                dim=-1,
            )
        )
        tau_ = tau_[..., 0]
        # assert switching.shape[-1] == alpha.shape[-1]
        # assert tau.shape == u_obs.shape == s_obs.shape == tau_.shape
        return softplus(switching), softplus(tau), softplus(tau_)


class AutoDeltaRNAVelocityGuide(autoguide.AutoDelta):
    def __init__(self, model: VelocityModel, use_gpu: int, **initial_values):
        if initial_values != {}:
            init_loc_fn = autoguide.init_to_value(
                values={
                    key: value.to(f"cuda:{use_gpu}")
                    if key != "cell_time"
                    else pyro.subsample(
                        value.to(f"cuda:{use_gpu}"), event_dim=0
                    )
                    for key, value in initial_values.items()
                }
            )
            # return autoguide.init_to_value(site, values={site["name"]: pyro.subsample(self.cell_time_init, event_dim=self.event_dim)})
        else:
            init_loc_fn = autoguide.init_to_median
        super().__init__(
            poutine.block(model, hide_fn=site_is_discrete),
            init_loc_fn=init_loc_fn,
            create_plates=model.create_plates,
        )


class AutoNormalRNAVelocityGuide(autoguide.AutoNormal):
    def __init__(self, model: VelocityModel, use_gpu: int, **initial_values):
        if initial_values != {}:
            init_loc_fn = autoguide.init_to_value(
                values={
                    key: value.to(f"cuda:{use_gpu}")
                    for key, value in initial_values.items()
                }
            )
        else:
            init_loc_fn = autoguide.init_to_median
        super().__init__(
            poutine.block(model, hide_fn=site_is_discrete),
            init_loc_fn=init_loc_fn,
            create_plates=model.create_plates,
        )


# class LatentGuide(easyguide.EasyGuide):
#     def __init__(
#         self,
#         model: LatentFactor,
#         latent_factor_size: int = 10,
#         plate_size: int = 2,
#         inducing_point_size: int = 20,
#         **initial_values,
#     ):
#         super().__init__(model)
#         self.num_genes = model.num_genes
#         self.num_cells = model.num_cells
#         self.zero = model.zero
#         self.mask = initial_values.get("mask", None)
#         self.latent_factor_size = latent_factor_size
#         self.plate_size = plate_size
#         self.inducing_point_size = inducing_point_size

#         for key in initial_values:
#             self.register_buffer(f"{key}_init", initial_values[key])

#     def init(self, site):
#         if site["name"] == "cell_code":
#             return autoguide.init_to_value(
#                 site,
#                 values={site["name"]: pyro.subsample(self.cell_code_init, event_dim=1)},
#             )
#         if hasattr(self, f"{site['name']}_init"):
#             return autoguide.init_to_value(
#                 site, values={site["name"]: getattr(self, f"{site['name']}_init")}
#             )
#         return super().init(site)

#     def guide(
#         self,
#         u_obs: Optional[torch.Tensor] = None,
#         s_obs: Optional[torch.Tensor] = None,
#         u_log_library: Optional[torch.Tensor] = None,
#         s_log_library: Optional[torch.Tensor] = None,
#         ind_x: Optional[torch.Tensor] = None,
#     ):
#         """max plate = 2 cell plate and gene plate"""
#         if self.plate_size == 2:
#             decoder_weights = self.map_estimate("cell_codebook")
#             encoder_weights = decoder_weights.T
#             gene_plate = self.plate("genes", self.num_genes, dim=-1)
#             with gene_plate:
#                 u_scale = self.map_estimate("u_scale")
#                 s_scale = self.map_estimate("s_scale")
#                 u_pcs_mean = self.map_estimate("u_pcs_mean")
#                 s_pcs_mean = self.map_estimate("s_pcs_mean")

#             if self.inducing_point_size == 0:
#                 cell_code_scale = pyro.param(
#                     "cell_code_scale", lambda: torch.tensor(0.1), constraint=positive
#                 ).to(u_obs.device)
#             cell_plate = self.plate("cells", self.num_cells, dim=-2, subsample=ind_x)
#             with cell_plate:
#                 cell_code_loc = (
#                     torch.cat((u_obs - u_pcs_mean, s_obs - s_pcs_mean), dim=-1)
#                     @ encoder_weights
#                 )
#                 if self.inducing_point_size > 0:
#                     inducing_points = pyro.param(
#                         "inducing_points",
#                         lambda: torch.randn(
#                             self.inducing_point_size, self.latent_factor_size
#                         ),
#                     ).to(u_obs.device)
#                     inducing_mean = pyro.param(
#                         "inducing_mean",
#                         lambda: torch.randn(
#                             self.inducing_point_size, self.latent_factor_size
#                         ),
#                     ).to(u_obs.device)

#                     kernel = pyro.contrib.gp.kernels.RBF(
#                         input_dim=self.latent_factor_size
#                     )
#                     assert cell_code_loc.shape == (
#                         cell_plate.subsample_size,
#                         self.latent_factor_size,
#                     )
#                     assert inducing_points.shape == (
#                         self.inducing_point_size,
#                         self.latent_factor_size,
#                     )
#                     assert inducing_mean.shape == (
#                         self.inducing_point_size,
#                         self.latent_factor_size,
#                     )
#                     cell_code_loc, cell_code_scale = pyro.contrib.gp.util.conditional(
#                         cell_code_loc,
#                         inducing_points,
#                         kernel,
#                         inducing_mean.transpose(-1, -2),
#                         full_cov=False,
#                     )
#                     assert cell_code_loc.shape == (
#                         self.latent_factor_size,
#                         cell_plate.subsample_size,
#                     )
#                     cell_code_loc = cell_code_loc.T.unsqueeze(-1).transpose(-1, -2)
#                     cell_code_scale = cell_code_scale.T.unsqueeze(-1).transpose(-1, -2)
#                 else:
#                     cell_code_loc = cell_code_loc.unsqueeze(-1).transpose(-1, -2)
#                 cell_code = pyro.sample(
#                     "cell_code", Normal(cell_code_loc, cell_code_scale).to_event(1)
#                 )
#         else:
#             cell_code = self.guide2(u_obs, s_obs, u_log_library, s_log_library, ind_x)
#         return cell_code

#     def guide2(
#         self,
#         u_obs: Optional[torch.Tensor] = None,
#         s_obs: Optional[torch.Tensor] = None,
#         u_log_library: Optional[torch.Tensor] = None,
#         s_log_library: Optional[torch.Tensor] = None,
#         ind_x: Optional[torch.Tensor] = None,
#     ):
#         """max plate = 1 only cell plate"""
#         cell_plate = self.plate("cells", self.num_cells, subsample=ind_x, dim=-1)
#         decoder_weights = self.map_estimate("cell_codebook")
#         encoder_weights = decoder_weights.T
#         u_scale = self.map_estimate("u_scale")
#         s_scale = self.map_estimate("s_scale")
#         u_pcs_mean = self.map_estimate("u_pcs_mean")
#         s_pcs_mean = self.map_estimate("s_pcs_mean")
#         cell_plate = self.plate("cells", self.num_cells, subsample=ind_x, dim=-1)
#         # cell_code_scale = pyro.param("cell_code_scale", lambda: torch.ones(self.latent_factor_size) * 0.1, constraint=positive, event_dim=1).to(u_obs.device)
#         cell_code_scale = pyro.param(
#             "cell_code_scale", lambda: torch.tensor(0.1), constraint=positive
#         ).to(u_obs.device)
#         with cell_plate:
#             cell_code = (
#                 torch.cat((u_obs - u_pcs_mean, s_obs - s_pcs_mean), dim=-1)
#                 @ encoder_weights
#             )
#             cell_code = pyro.sample(
#                 "cell_code", Normal(cell_code, cell_code_scale).to_event(1)
#             )
#         return cell_code
